import logging

import requests
from bs4 import BeautifulSoup
from pythoncommons.file_utils import FileUtils
from pythoncommons.string_utils import StringUtils

LOG = logging.getLogger(__name__)


class JiraUtils:
    @staticmethod
    def download_jira_html(jira_server, jira_id, to_file):
        resp = requests.get(jira_server + jira_id)
        resp.raise_for_status()
        FileUtils.save_to_file(to_file, resp.text)
        return resp.text

    @staticmethod
    def parse_jira_title(html_doc):
        soup = BeautifulSoup(html_doc, "html.parser")
        title = soup.find("h1", attrs={"id": "summary-val"})
        if not title:
            raise ValueError("Cannot find title in HTML document: {}".format(html_doc))
        return title.string

    @staticmethod
    def parse_subjiras_from_umbrella_html(html_doc, to_file, filter_ids):
        soup = BeautifulSoup(html_doc, "html.parser")
        issue_keys = []
        for link in soup.find_all("a", attrs={"class": "issue-link"}):
            issue_keys.append(link.attrs["data-issue-key"])

        if filter_ids:
            LOG.info("Filtering ids from result list: %s", filter_ids)
            issue_keys = [issue for issue in issue_keys if issue not in filter_ids]

        # Filter dupes
        issue_keys = list(set(issue_keys))
        FileUtils.save_to_file(to_file, StringUtils.list_to_multiline_string(issue_keys))
        return issue_keys

    @staticmethod
    def parse_subjiras_and_jira_titles_from_umbrella_html(html_doc, to_file, filter_ids, find_all_links=True):
        soup = BeautifulSoup(html_doc, "html.parser")
        result_dict = {}

        links = []
        if find_all_links:
            links = soup.find_all("a", attrs={"class": "issue-link"})
        else:
            table = soup.find("table", id="issuetable")
            if table is not None:
                links = table.find_all("a", attrs={"class": "issue-link"})
        for link in links:
            jira_id = link.attrs["data-issue-key"]
            jira_title = str(link.contents[0])
            # There are 2 anchors with class: 'issue-link' per row. Only store the one with valid title.
            if len(jira_title.strip()) > 0:
                result_dict[jira_id] = jira_title

        if filter_ids:
            LOG.info("Filtering ids from result list: %s", filter_ids)
            result_dict = {jira_id: title for jira_id, title in result_dict.items() if jira_id not in filter_ids}

        FileUtils.save_to_file(to_file, StringUtils.dict_to_multiline_string(result_dict))
        return result_dict



import logging
import os

from pythoncommons.file_utils import FileUtils
from pythoncommons.string_utils import StringUtils

PATCH_FILE_SEPARATOR = "."
REVIEW_BRANCH_SEP = "-"
FIRST_PATCH_NUMBER = "001"
LOG = logging.getLogger(__name__)


class PatchUtils:
    @staticmethod
    def extract_patch_number_from_filename_as_int(filename, pos=-2):
        # Assuming filename like: '/somedir/YARN-10277-test.0003.patch'
        return int(filename.split(PATCH_FILE_SEPARATOR)[pos])

    @staticmethod
    def extract_patch_number_from_filename_as_str(filename, pos=-2):
        # Assuming filename like: '/somedir/YARN-10277-test.0003.patch'
        return filename.split(PATCH_FILE_SEPARATOR)[pos]

    @staticmethod
    def get_next_patch_filename(filename, pos=-2):
        # Assuming filename like: '/somedir/YARN-10277-test.0003.patch'
        split = filename.split(PATCH_FILE_SEPARATOR)
        increased_str = StringUtils.increase_numerical_str(split[pos])
        split[pos] = increased_str
        return PATCH_FILE_SEPARATOR.join(split)

    @staticmethod
    def get_next_filename(patch_dir, list_of_prev_patches):
        list_of_prev_patches = sorted(list_of_prev_patches, reverse=True)
        LOG.info("Found patches: %s", list_of_prev_patches)
        if len(list_of_prev_patches) == 0:
            return FileUtils.join_path(patch_dir, FIRST_PATCH_NUMBER), FIRST_PATCH_NUMBER
        else:
            latest_patch = list_of_prev_patches[0]
            last_patch_num = PatchUtils.extract_patch_number_from_filename_as_str(latest_patch)
            next_patch_filename = PatchUtils.get_next_patch_filename(latest_patch)
            return (
                FileUtils.join_path(patch_dir, next_patch_filename),
                StringUtils.increase_numerical_str(last_patch_num),
            )

    @staticmethod
    def get_next_review_branch_name(branches, sep=REVIEW_BRANCH_SEP):
        # review-YARN-10277-3
        # review-YARN-10277-2
        # review-YARN-10277
        sorted_branches = sorted(branches, reverse=True)
        if len(sorted_branches) == 0:
            raise ValueError("Expected a list of branches with size 1 at least. List: %s", sorted_branches)

        latest_branch = sorted_branches[0]
        parts = latest_branch.split(sep)

        if len(parts) < 3:
            raise ValueError(
                "Expected at least 3 components (separated by '-') of branch name: {}, encountered: {}",
                latest_branch,
                len(parts),
            )

        # No branch postfix, e.g. review-YARN-10277
        if len(parts) == 3:
            return sep.join(parts) + sep + "2"
        elif len(parts) == 4:
            return sep.join(parts[0:3]) + sep + StringUtils.increase_numerical_str(parts[3])
        else:
            raise ValueError(
                "Unexpected number of components (separated by '-') of branch name: {}, "
                "encountered # of components: {}", latest_branch, len(parts))

    @staticmethod
    def save_diff_to_patch_file(diff, file):
        if not diff or diff == "":
            LOG.error("Diff was empty. Patch file is not created!")
            return False
        else:
            diff += os.linesep
            LOG.info("Saving diff to patch file: %s", file)
            LOG.debug("Diff: %s", diff)
            FileUtils.save_to_file(file, diff)
            return True


import logging
import os
import site
import sys

from pythoncommons.project_utils import ProjectUtils, ProjectRootDeterminationStrategy
from testproject.commands.testcommand.dummy_test_command import FakeBranchComparator

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
LOG = logging.getLogger("project_utils")
LOG.setLevel(logging.DEBUG)


def main():
    LOG = logging.getLogger(__name__)
    ProjectUtils.project_root_determine_strategy = ProjectRootDeterminationStrategy.SYS_PATH
    output_basedir = ProjectUtils.get_output_basedir("testproject")
    output_child_dir = ProjectUtils.get_output_child_dir("branch-comparator")
    logfilename = ProjectUtils.get_default_log_file("test-log")

    LOG.info("Output basedir: " + output_basedir)
    LOG.info("Output child dir: " + output_child_dir)
    LOG.info("logfilename: " + logfilename)

    python_global_site = site.getsitepackages()[0]
    python_user_site = site.USER_SITE
    LOG.info("sys.path: %s", sys.path)
    LOG.info("Global site: " + python_global_site)
    LOG.info("User site: " + python_user_site)

    LOG.info("Creating FakeBranchComparator...")
    FakeBranchComparator()
    ProjectUtils.reset_root_determine_strategy_to_default()


if __name__ == "__main__":
    main()



import json
import logging
import random
import urllib
import urllib.request as request
import urllib.error as url_error
from typing import Dict, Set, Callable

LOG = logging.getLogger(__name__)


class NetworkUtils:
    @staticmethod
    def is_port_in_use(port):
        import socket
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex(('localhost', port)) == 0

    @staticmethod
    def get_random_port():
        return random.randrange(5000, 6000)

    @staticmethod
    def wait_for_internet_connection():
        counter = 0
        while True:
            try:
                # ping google
                if counter > 0:
                    LOG.info('Waiting for internet connection, try count: %s', counter)
                response = request.urlopen('http://google.com', timeout=1)
                return
            except url_error.URLError:
                pass

    @staticmethod
    def fetch_json(url, strict=False,
                   do_not_raise_http_statuses: Set[int] = None,
                   http_callbacks: Dict[int, Callable] = None):
        """ Load data from specified url """
        try:
            LOG.debug("Making request to URL: %s", url)
            ourl = urllib.request.urlopen(url)
            codec = ourl.info().get_param("charset")
            content = ourl.read().decode(codec)
        except urllib.error.HTTPError as e:
            if do_not_raise_http_statuses and e.code in do_not_raise_http_statuses:
                if http_callbacks and e.code in http_callbacks:
                    http_callbacks[e.code]()
                return {}
            else:
                raise e
        return json.loads(content, strict=strict)



import logging
import os
from contextlib import closing
from io import BufferedWriter
from typing import List
from zlib import Z_DEFAULT_COMPRESSION

from pythoncommons.file_utils import FileUtils, FileFinder
import tempfile
import zipfile

from pythoncommons.string_utils import StringUtils

LOG = logging.getLogger(__name__)


class ZipFileUtils:
    @staticmethod
    def create_zip_file_advanced(
        input_files: List[str], dest_filename: str, ignore_filetypes: List[str] = None, output_dir: str = None
    ):
        if not ignore_filetypes:
            ignore_filetypes = []

        sum_len_all_files: int = 0
        all_ignores_files: int = 0
        if ignore_filetypes:
            (
                all_ignores_files,
                sum_len_all_files,
                input_files,
                tmp_dir,
            ) = ZipFileUtils._determine_input_files_based_on_exclusions(
                all_ignores_files, ignore_filetypes, input_files, sum_len_all_files
            )

        temp_dir_dest: bool = True if not output_dir or output_dir.startswith("/tmp") else False
        if output_dir:
            dest_filepath = FileUtils.join_path(output_dir, dest_filename)
            zip_file: BufferedWriter = ZipFileUtils.create_zip_file(input_files, dest_filepath, compress=True)
        else:
            zip_file: BufferedWriter = ZipFileUtils.create_zip_as_tmp_file(input_files, dest_filename, compress=True)

        zip_file_name = zip_file.name
        no_of_files_in_zip: int = ZipFileUtils.get_number_of_files_in_zip(zip_file_name)
        if ignore_filetypes and (sum_len_all_files - all_ignores_files) != no_of_files_in_zip:
            raise ValueError(
                f"Unexpected number of files in zip. "
                f"All files: {sum_len_all_files}, "
                f"all ignored files: {all_ignores_files}, "
                f"number of files in zip: {no_of_files_in_zip}, "
                f"zip file: {zip_file_name}"
            )

        LOG.info(
            f"Finished writing command data to zip file: {zip_file_name}, "
            f"size: {FileUtils.get_file_size(zip_file_name)}"
        )
        return zip_file_name, temp_dir_dest

    @staticmethod
    def _determine_input_files_based_on_exclusions(
        all_ignores_files, ignore_filetypes, input_files_param, sum_len_all_files
    ):
        input_files = []
        tmp_dir: tempfile.TemporaryDirectory or None = None
        for input_file in input_files_param:
            if FileUtils.is_dir(input_file):
                all_files = FileUtils.find_files(input_file, regex=".*", full_path_result=True)
                sum_len_all_files += len(all_files)
                files_to_ignore = set()
                for ext in ignore_filetypes:
                    new_files_to_ignore = FileUtils.find_files(input_file, extension=ext, full_path_result=True)
                    all_ignores_files += len(new_files_to_ignore)
                    LOG.debug(
                        f"Found {len(new_files_to_ignore)} files to ignore in directory '{input_file}': "
                        f"{StringUtils.list_to_multiline_string(files_to_ignore)}"
                    )
                    files_to_ignore.update(new_files_to_ignore)

                files_to_keep = list(set(all_files).difference(files_to_ignore))
                tmp_dir: tempfile.TemporaryDirectory = tempfile.TemporaryDirectory()
                tmp_dir_path = tmp_dir.name
                FileUtils.copy_files_to_dir(files_to_keep, tmp_dir_path, cut_path=input_file)
                input_files.append(tmp_dir_path)
            else:
                input_files.append(input_file)
                sum_len_all_files += 1
        # IMPORTANT: Need to return tmp_dir as when the TemporaryDirectory object is garbage collected, the tmp dir itself is deleted as well
        # See: https://stackoverflow.com/a/55104228/1106893
        return all_ignores_files, sum_len_all_files, input_files, tmp_dir

    @staticmethod
    def create_zip_as_tmp_file(src_files: List[str], filename: str, compress=False):
        filename, suffix = ZipFileUtils._validate_zip_file_name(filename)
        tmp_file = tempfile.NamedTemporaryFile(prefix=filename, suffix=suffix, delete=False)
        return ZipFileUtils._create_zip_file(src_files, tmp_file, compress=compress)

    @staticmethod
    def create_zip_file(src_files: List[str], filename: str, compress=False, ignore_files: List[str] = None):
        return ZipFileUtils._create_zip_file(
            src_files, open(filename, mode="wb"), compress=compress, ignore_files=ignore_files
        )

    @staticmethod
    def extract_zip_file(file: str, path: str):
        # Apparently, ZipFile does not resolve symlinks so let's do it manually
        if os.path.islink(file):
            file = os.path.realpath(file)
        FileUtils.ensure_file_exists(file)
        zip_file = zipfile.ZipFile(file, "r")
        zip_file.extractall(path)

    @staticmethod
    def _validate_zip_file_name(filename):
        if "." in filename:
            filename_and_ext = filename.split(".")
            if len(filename_and_ext) != 2:
                raise ValueError("Invalid filename: " + filename)
            filename = filename_and_ext[0]
            suffix = "." + filename_and_ext[1]
        else:
            filename = filename
            suffix = ".zip"
        return filename, suffix

    @staticmethod
    def _create_zip_file(src_files, file, compress=False, ignore_files: List[str] = None):
        kwargs = {}
        if compress:
            kwargs["compression"] = zipfile.ZIP_DEFLATED
            kwargs["compresslevel"] = Z_DEFAULT_COMPRESSION  # https://docs.python.org/3/library/zlib.html#zlib.compress
        if not ignore_files:
            ignore_files = []
        zip_file = zipfile.ZipFile(file, "w", **kwargs)
        LOG.info(f"Creating zip file. Target file: {zip_file.filename}, Input files: {src_files}")
        for src_file in src_files:
            if not FileUtils.does_file_exist(src_file):
                LOG.warning("Src file does not exist: %s", src_file)
                continue
            if ignore_files:
                if ZipFileUtils._is_file_ignored(src_file, ignore_files):
                    continue
            if FileUtils.is_dir(src_file, throw_ex=False):
                ZipFileUtils._add_dir_to_zip(src_file, zip_file, ignore_files=ignore_files)
            else:
                LOG.debug(f"Adding file '{src_file}' to zip file '${zip_file.filename}'")
                zip_file.write(src_file, FileUtils.basename(src_file))
        zip_file.close()
        file.seek(0)
        return file

    @staticmethod
    def _is_file_ignored(input_file, ignore_files):
        for ignore in ignore_files:
            if ignore in input_file:
                LOG.debug("Ignoring file while zipping: %s", input_file)
                return True
        return False

    # TODO duplicated os.walk code from FileFinder --> Migrate
    @staticmethod
    def _add_dir_to_zip(src_dir, zip_file, ignore_files: List[str] = None):
        # Iterate over all the files in directory
        LOG.debug(f"Adding directory '{src_dir}' to zip file '${zip_file.filename}'")
        for dirpath, dirnames, filenames in os.walk(src_dir, **FileFinder._get_os_walk_kwargs(ignore_files)):
            LOG.debug(f"[os.walk] dirpath: {dirpath}, dirnames: {dirnames}, filenames: {filenames}")
            dirnames[:] = ZipFileUtils._handle_dir_exclusions(dirnames, ignore_files)
            for filename in filenames:
                ZipFileUtils._add_file_to_zip(zip_file, dirpath, filename, src_dir)

    # TODO duplicated code from FileFinder --> Migrate
    @classmethod
    def _handle_dir_exclusions(cls, dirs, exclude_files):
        if exclude_files:
            # Not enough to check against basename(root) as all other dirs underneath will be walked on the next
            # invocation of the walk generator with the loop statement
            orig_dirs = dirs.copy()
            dirs[:] = [d for d in dirs if d not in exclude_files]
            if len(orig_dirs) != len(dirs):
                LOG.debug(f"Excluded dirs: {list(set(orig_dirs) - set(dirs))}")
        return dirs

    @staticmethod
    def _add_file_to_zip(zip, dirpath, filename, src_dir):
        file_full_path = os.path.join(dirpath, filename)
        dir_path_from_src_dir = dirpath.replace(src_dir, "")
        if dir_path_from_src_dir.startswith(os.sep):
            dir_path_from_src_dir = dir_path_from_src_dir[1:]

        path_in_zip = FileUtils.join_path(dir_path_from_src_dir, filename)
        LOG.debug(f"Writing to zip file {zip}. File full path: {file_full_path}, path in zip file: {path_in_zip}")
        zip.write(file_full_path, path_in_zip)

    @staticmethod
    def get_number_of_files_in_zip(zip_file: str):
        with closing(zipfile.ZipFile(zip_file)) as archive:
            count = len(archive.infolist())
        return count



import unittest
import logging
import sys

from pythoncommons.logging_utils import LoggerFactory, COLLECTION_PLACEHOLDER

TESTLOGGER_INFO = "INFO:testLogger:"
TESTLOGGER_DEBUG = "DEBUG:testLogger:"


def double_coll(coll):
    coll *= 2
    return f"{str(coll)}"


class LoggingUtilsTests(unittest.TestCase):

    def test_many_strings(self):
        # format="%(message)s"
        logging.basicConfig(stream=sys.stdout, level=logging.WARN)

        with self.assertLogs('testLogger', level='INFO') as cm:
            logger = LoggerFactory.get_logger("testLogger")
            logger.setLevel(logging.WARN)
            logger.combined_log("testInfo0", ["a1", "a2"])

            logger.setLevel(logging.INFO)
            logger.combined_log("testInfo1...|", ["a1", "a2"])
            logger.combined_log(f"te {COLLECTION_PLACEHOLDER} stInfo2...", ["a1", "a2", "a3"])

            logger.setLevel(logging.DEBUG)
            logger.combined_log("testDebug1...", ["a1", "a2"])
            logger.combined_log(f"te {COLLECTION_PLACEHOLDER} stDebug2...", ["a1", "a2", "a3"])
            logger.combined_log(f"te {COLLECTION_PLACEHOLDER} stDebug3...", ["a1", "a2", "a3", "a4"], debug_coll_func=double_coll)

            self.assertEqual(cm.output, [self.logger_info_msg("testInfo1...| 2"),
                                         self.logger_info_msg("te 3 stInfo2..."),

                                         self.logger_info_msg("testDebug1... 2"),
                                         self.logger_debug_msg("testDebug1... ['a1', 'a2']"),

                                         self.logger_info_msg("te 3 stDebug2..."),
                                         self.logger_debug_msg("te ['a1', 'a2', 'a3'] stDebug2..."),

                                         self.logger_info_msg("te 4 stDebug3..."),
                                         self.logger_debug_msg("te ['a1', 'a2', 'a3', 'a4', "
                                                                   "'a1', 'a2', 'a3', 'a4'] stDebug3...")
                                         ])

    def logger_info_msg(self, msg):
        return f"{TESTLOGGER_INFO}{msg}"

    def logger_debug_msg(self, msg):
        return f"{TESTLOGGER_DEBUG}{msg}"










import secrets
import string


class RandomUtils:
    @classmethod
    def get_secure_random_string(cls, length):
        secure_str = ''.join((secrets.choice(string.ascii_letters) for i in range(length)))
        return secure_str



import logging
import os
import sys

# TODO rename to SystemUtils?
from typing import Dict, List

PASSWORD_PREFIX = "password "
LOG = logging.getLogger(__name__)


class OsUtils:
    ENV_UPDATES: Dict[str, str] = {}
    TRACK_UPDATES = False

    @classmethod
    def track_env_updates(cls):
        cls.TRACK_UPDATES = True

    @classmethod
    def stop_tracking_updates(cls, clear_updates_dict=False):
        cls.TRACK_UPDATES = False
        if clear_updates_dict:
            cls.ENV_UPDATES.clear()

    @classmethod
    def get_tracked_updates(cls) -> Dict[str, str]:
        return dict(cls.ENV_UPDATES)

    @classmethod
    def get_env_value(cls, env_name, default_value=None, suppress=False):
        if env_name in os.environ:
            env_value = os.environ[env_name]
            if not suppress:
                LOG.debug("Value of env variable '%s': %s", env_name, env_value)
        else:
            env_value = default_value
            if not suppress:
                LOG.debug("Value of env variable '%s' is not defined, using default value of: %s", env_name, env_value)
        return env_value

    @classmethod
    def set_env_value(cls, env_name, env_value, suppress=False):
        if env_name in os.environ:
            old_value = os.environ[env_name]
            if not suppress:
                LOG.debug(
                    "Setting value of env variable '%s' to : %s. Old value was: %s", env_name, env_value, old_value
                )
        else:
            if isinstance(env_value, bool):
                env_value = str(env_value)
            if not suppress:
                LOG.debug("Setting value of env variable '%s' to : %s", env_name, env_value)
        if cls.TRACK_UPDATES:
            cls.ENV_UPDATES[env_name] = env_value
        os.environ[env_name] = env_value

    @classmethod
    def clear_env_vars(cls, vars: List[str]):
        for var in vars:
            if var in os.environ:
                LOG.debug("Removing env variable '%s'", var)
                del os.environ[var]

    @classmethod
    def is_env_var_true(cls, env_var_name, default_val):
        env_var_value = OsUtils.get_env_value(env_var_name, default_val)
        if env_var_value is None:
            raise ValueError("Env var value should not be None for env var name: '{}'".format(env_var_name))
        return env_var_value == "True" or env_var_value is True

    @staticmethod
    def determine_full_command():
        return " ".join(sys.argv)

    @staticmethod
    def determine_full_command_filtered(filter_password=False):
        full_command = " ".join(sys.argv)
        if filter_password:
            split_res = full_command.split(PASSWORD_PREFIX)
            if len(split_res) == 1:
                # Password not found, return full command
                return full_command
            # Chop the first word from the 2nd string, that word should be the password.
            return split_res[0] + f"{PASSWORD_PREFIX}****** " + " ".join(split_res[1].split(" ")[1:])
        else:
            return full_command


class EnvironmentValidator:
    @staticmethod
    def validate_env_vars(env_vars, action=sys.exit, action_params=1, logger=LOG):
        for env_var in env_vars:
            if not os.getenv(env_var):
                logger.error("env var '%s' is not defined! Calling action: '%s(%s)'", env_var, action, action_params)
                action(action_params)



import logging
from email.mime.text import MIMEText
import smtplib
from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from enum import Enum
from typing import List

from pythoncommons.file_utils import FileUtils

LOG = logging.getLogger(__name__)


class EmailMimeType(Enum):
    HTML = "html"
    PLAIN = "plain"


class EmailAccount:
    def __init__(self, user: str, password: str):
        self.user: str = user
        self.password: str = password


class EmailConfig:
    def __init__(self, smtp_server: str, smtp_port: int, email_account: EmailAccount):
        self.smtp_server = smtp_server
        self.smtp_port = smtp_port
        self.email_account = email_account


class EmailService:
    def __init__(self, email_config: EmailConfig):
        self.conf = email_config

    def send_mail(
        self,
        sender: str,
        subject: str,
        body: str,
        recipients: List[str],
        attachment_file=None,
        override_attachment_filename: str = None,
        body_mimetype: EmailMimeType = EmailMimeType.PLAIN,
        with_retries: bool = True,
        retry_count: int = 3,
        log_exception_when_retried: bool = True,
    ):
        self._validate_config(recipients)
        saved_args = locals().copy()
        # Do not log HTML contents
        del saved_args["body"]
        LOG.debug("Received args: %s", saved_args)
        mime_text = MIMEText(str(body), body_mimetype.value)
        if not attachment_file:
            email_msg = mime_text
        else:
            FileUtils.ensure_file_exists(attachment_file)
            # https://stackoverflow.com/a/169406/1106893
            email_msg = MIMEMultipart()
            email_msg.attach(mime_text)
            if override_attachment_filename:
                attachment = self._create_attachment(attachment_file, attachment_name=override_attachment_filename)
            else:
                attachment = self._create_attachment(attachment_file)
            email_msg.attach(attachment)
        recipients_comma_separated = self._add_common_email_data(email_msg, recipients, sender, subject)
        self._connect_to_server_and_send(
            email_msg,
            recipients,
            recipients_comma_separated,
            sender,
            with_retries=with_retries,
            retry_count=retry_count,
            log_exception_when_retried=log_exception_when_retried,
        )

    @staticmethod
    def _add_common_email_data(email_msg, recipients, sender, subject):
        recipients_comma_separated = ", ".join(recipients)
        email_msg["From"] = sender
        email_msg["To"] = recipients_comma_separated
        email_msg["Subject"] = subject
        email_msg.preamble = "I am not using a MIME-aware mail reader.\n"
        return recipients_comma_separated

    def _validate_config(self, recipients):
        if not recipients:
            LOG.error("Cannot send email as recipient email addresses are not set!")
            return
        if not self.conf:
            raise ValueError("Email config is not set!")
        if not all(attr is not None for attr in vars(self.conf)):
            raise ValueError(f"Some attribute of EmailConfig is not set. Config object: {self.conf}")
        if not self.conf.email_account.user:
            raise ValueError("Wrong email server config. Username must be set!")
        if not self.conf.email_account.password:
            raise ValueError("Wrong email server config. Password must be set!")

    def _connect_to_server_and_send(
        self,
        email_msg,
        recipients,
        recipients_comma_separated,
        sender,
        with_retries: bool = True,
        retry_count: int = 3,
        log_exception_when_retried: bool = True,
    ):
        server = smtplib.SMTP_SSL(self.conf.smtp_server, self.conf.smtp_port)
        if with_retries:
            all_try_count: int = retry_count + 1
        else:
            all_try_count: int = 1

        for i in range(all_try_count):
            attempt = i + 1
            LOG.info(
                "[Attempt: %d / %d] Sending mail to recipients: %s with subject '%s'",
                attempt,
                all_try_count,
                recipients_comma_separated,
                email_msg["Subject"],
            )
            try:
                server.ehlo()
                server.login(self.conf.email_account.user, self.conf.email_account.password)
                server.sendmail(sender, recipients, email_msg.as_string())
                server.quit()
                return
            except smtplib.SMTPServerDisconnected as e:
                if attempt == all_try_count:
                    # Raise if we reached max retries or we just tried once without retries and it failed
                    raise e
                elif log_exception_when_retried:
                    LOG.exception("Failed to send email.", exc_info=True)

    @staticmethod
    def _create_attachment(file_path: str, attachment_name: str = None):
        msg = MIMEBase("application", "zip")
        file = open(file_path, "rb")
        msg.set_payload(file.read())
        encoders.encode_base64(msg)

        attachment_fname_header = EmailService._get_attachment_name(attachment_name, file_path)
        msg.add_header("Content-Disposition", "attachment", filename=attachment_fname_header)
        return msg

    @staticmethod
    def _get_attachment_name(attachment_name, file_path):
        name = FileUtils.basename(file_path)
        if attachment_name:
            name: str = attachment_name
        if not name.endswith(".zip"):
            name += ".zip"
        return name



import logging
import sys

from pythoncommons.project_utils import ProjectUtils, ProjectRootDeterminationStrategy

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
LOG = logging.getLogger("project_utils")
LOG.setLevel(logging.DEBUG)


def main():
    LOG = logging.getLogger(__name__)
    LOG.info("hello world")

    LOG.info("sys.path: %s", sys.path)
    ProjectUtils.project_root_determine_strategy = ProjectRootDeterminationStrategy.SYS_PATH
    ProjectUtils.FORCE_SITE_PACKAGES_IN_PATH_NAME = False
    basedir = ProjectUtils.get_output_basedir("test")
    logfilename = ProjectUtils.get_default_log_file("test")
    LOG.info("basedir: " + basedir)
    LOG.info("logfilename: " + logfilename)

    ProjectUtils.reset_root_determine_strategy_to_default()


if __name__ == "__main__":
    main()



import datetime
import logging
import time

LOG = logging.getLogger(__name__)

DATEFORMAT_DASH_COLON = "%Y-%m-%d %H:%M:%S"


# https://medium.com/pythonhive/python-decorator-to-measure-the-execution-time-of-methods-fa04cb6bb36d
def timeit(method):
    def timed(*args, **kw):
        def print_time(method_name, time_value):
            pretty_time = TimeUtilities.prettify_time(time_value)
            LOG.info("Method runtime: %s: %s", method_name, pretty_time)

        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if "log_time" in kw:
            name = kw.get("log_name", method.__name__.upper())
            kw["log_time"][name] = int((te - ts) * 1000)
        else:
            print_time(method.__name__, (te - ts))
        return result

    return timed


class TimeUtilities:
    SECONDS = "seconds"
    MINUTES = "minutes"
    HOURS = "hours"
    DAYS = "days"

    @classmethod
    def prettify_time(cls, seconds):
        # Credits: https://gist.github.com/thatalextaylor/7408395
        seconds = int(seconds)
        days, seconds = divmod(seconds, 86400)
        hours, seconds = divmod(seconds, 3600)
        minutes, seconds = divmod(seconds, 60)

        # TODO generate pairs of curly braces with range()
        if days > 0:
            return "{} {}, {} {}, {} {}, {} {}".format(
                days, cls.DAYS, hours, cls.HOURS, minutes, cls.MINUTES, seconds, cls.SECONDS
            )
        elif hours > 0:
            return "{} {}, {} {}, {} {}".format(hours, cls.HOURS, minutes, cls.MINUTES, seconds, cls.SECONDS)
        elif minutes > 0:
            return "{} {}, {} {}".format(minutes, cls.MINUTES, seconds, cls.SECONDS)
        else:
            return "{} {}".format(seconds, cls.SECONDS)


class DateUtils:
    WIN_EPOCH = datetime.datetime(1601, 1, 1)

    @staticmethod
    def get_current_datetime(fmt="%Y%m%d_%H%M%S"):
        return DateUtils.now_formatted(fmt)

    @classmethod
    def now_formatted(cls, fmt):
        return DateUtils.now().strftime(fmt)

    @classmethod
    def from_iso_format(cls, date_str):
        return datetime.date.fromisoformat(date_str)

    @classmethod
    def now(cls):
        return datetime.datetime.now()

    @classmethod
    def create_datetime_from_timestamp(cls, timestamp: int):
        return datetime.datetime.fromtimestamp(timestamp)

    @classmethod
    def format_unix_timestamp(cls, timestamp: int):
        dt = DateUtils.create_datetime_from_timestamp(timestamp)
        return DateUtils.convert_datetime_to_str(dt, DATEFORMAT_DASH_COLON)

    @classmethod
    def add_microseconds_to_win_epoch(cls, microseconds):
        return DateUtils.WIN_EPOCH + datetime.timedelta(microseconds=microseconds)

    @classmethod
    def get_current_time_minus(cls, seconds=None, minutes=None, hours=None, days=None) -> datetime:
        now = datetime.datetime.now()
        return cls._datetime_minus(now, days, hours, minutes, seconds)

    @classmethod
    def datetime_minus(cls, dt: datetime, seconds=None, minutes=None, hours=None, days=None) -> datetime:
        return cls._datetime_minus(dt, days, hours, minutes, seconds)

    @classmethod
    def _datetime_minus(cls, dt: datetime, days, hours, minutes, seconds) -> datetime:
        kwargs = {}
        if seconds:
            kwargs["seconds"] = seconds
        if minutes:
            kwargs["minutes"] = minutes
        if hours:
            kwargs["hours"] = hours
        if days:
            kwargs["days"] = days
        return dt - datetime.timedelta(**kwargs)

    @classmethod
    def convert_to_datetime(cls, date_string, fmt):
        return datetime.datetime.strptime(date_string, fmt)

    @classmethod
    def convert_datetime_to_str(cls, datetime_obj, fmt):
        return datetime_obj.strftime(fmt)

    @classmethod
    def get_datetime(cls, y, m, d):
        return datetime.datetime(y, m, d)

    @classmethod
    def get_datetime_from_date(cls, date_obj, min_time=False, max_time=False):
        if not min_time and not max_time:
            raise ValueError("Either min_time or max_time had set to True!")

        if min_time:
            return datetime.datetime.combine(date_obj, datetime.datetime.min.time())
        if max_time:
            return datetime.datetime.combine(date_obj, datetime.datetime.max.time())

    @classmethod
    def reset_to_midnight(cls, dt: datetime.datetime):
        return datetime.datetime.combine(dt, datetime.datetime.min.time())



import logging
from dataclasses import dataclass
from enum import Enum
from typing import Tuple, List, Any
from colr import color
from pythoncommons.string_utils import StringUtils, auto_str
from tabulate import tabulate

LOG = logging.getLogger(__name__)


class TabulateTableFormat(Enum):
    GRID = "fancy_grid"
    HTML = "html"
    UNSAFE_HTML = "unsafehtml"


DEFAULT_TABLE_FORMATS = [TabulateTableFormat.GRID, TabulateTableFormat.HTML]


@dataclass
class TableCellLink:
    link_name: str
    link_value: str

    def to_html(self):
        return f'<a href="{self.link_value}">{self.link_name}</a>'


class GenericTableWithHeader:
    def __init__(
        self,
        header_title: str,
        header_list: List[str],
        source_data: Any,
        rendered_table: str,
        table_fmt: TabulateTableFormat,
        colorized: bool = False,
    ):
        self.header = (
            StringUtils.generate_header_line(
                header_title, char="═", length=len(StringUtils.get_first_line_of_multiline_str(rendered_table))
            )
            + "\n"
        )
        self.header_list = header_list
        self.source_data = source_data
        self.table = rendered_table
        self.table_fmt: TabulateTableFormat = table_fmt
        self.colorized = colorized

    def __str__(self):
        return self.header + self.table


class Color(Enum):
    GREEN = "green"
    RED = "red"


class ColorType(Enum):
    FOREGROUND = "fore"
    BACKROUND = "back"


class MatchType(Enum):
    ALL = "all"
    ANY = "any"


class EvaluationMethod(Enum):
    ALL = 0
    FIRST_TRUTHY = 1


@auto_str
class ColorDescriptor:
    def __init__(
        self,
        type,
        value,
        color: Color,
        match_type: MatchType,
        scan_range: Tuple[int, int],
        colorize_range: Tuple[int, int],
        color_type: ColorType = ColorType.FOREGROUND,
    ):
        self.type = type
        self.value = value
        self.color: Color = color
        self.match_type: MatchType = match_type
        self.scan_range: Tuple[int, int] = scan_range
        self.colorize_range: Tuple[int, int] = colorize_range
        self.color_type = color_type


@auto_str
class ConversionResult:
    def __init__(self, src_data, result_data):
        self.src_data = src_data
        self.dst_data = result_data


@auto_str
class ColorizeConfig:
    def __init__(
        self, descriptors: List[ColorDescriptor], eval_method: EvaluationMethod = EvaluationMethod.FIRST_TRUTHY
    ):
        self.descriptors = descriptors
        self.eval_method = eval_method


@auto_str
class BoolConversionConfig:
    def __init__(self, convert_true_to="X", convert_false_to="-"):
        self.convert_true_to = convert_true_to
        self.convert_false_to = convert_false_to


@auto_str
class TableRenderingConfig:
    def __init__(
        self,
        row_callback,  # TODO specify type
        join_lists_by_comma: bool = True,
        add_row_numbers: bool = True,
        max_width: int = None,
        max_width_separator: str = " ",
        bool_conversion_config: BoolConversionConfig = None,
        colorize_config: ColorizeConfig = None,
        print_result: bool = True,
        tabulate_format: TabulateTableFormat = None,
        tabulate_formats: List[TabulateTableFormat] = None,
    ):
        self.row_callback = row_callback
        self.join_lists_by_comma = join_lists_by_comma
        self.add_row_numbers = add_row_numbers
        self.max_width = max_width
        self.max_width_separator = max_width_separator
        self.bool_conversion_config = bool_conversion_config
        self.colorize_config = colorize_config
        self.print_result = print_result
        self.tabulate_format = tabulate_format
        self.tabulate_formats = tabulate_formats

        # Validation
        if not self.tabulate_format and not self.tabulate_formats:
            raise ValueError("Either of tabulate_format or tabulate_formats should be specified!")
        if self.tabulate_format and self.tabulate_formats:
            raise ValueError(
                "Can't decide if tabulate_format or tabulate_formats should be used. Please only use either of them!"
            )


class ResultPrinter:
    @staticmethod
    def print_tables(
        data,
        header,
        render_conf: TableRenderingConfig,
        verbose=False,
    ):
        tables = {}
        if verbose:
            LOG.debug("Rendering config for table is: %s", render_conf)
        for table_format in render_conf.tabulate_formats:
            # TODO this is a dirty hack to avoid: AttributeError: 'NoneType' object has no attribute 'value'
            render_conf.tabulate_format = table_format
            if verbose:
                LOG.debug("Calling %s with table format: %s", ResultPrinter.print_table.__name__, table_format)
            tables[table_format] = ResultPrinter.print_table(data, header, render_conf=render_conf)
        return tables

    @staticmethod
    def print_table(data, header, render_conf: TableRenderingConfig):
        conversion_result = ResultPrinter.convert_list_data(data, render_conf)
        # LOG.debug(f"Conversion result: {conversion_result}")
        tabulated = tabulate(conversion_result.dst_data, header, tablefmt=render_conf.tabulate_format.value)
        if render_conf.print_result:
            print(tabulated)
        return tabulated

    @staticmethod
    def convert_list_data(src_data, render_conf: TableRenderingConfig):
        result = []
        for idx, src_row in enumerate(src_data):
            row = render_conf.row_callback(src_row)
            converted_row = []
            if render_conf.add_row_numbers:
                converted_row.append(idx + 1)
            for cell in row:
                if render_conf.join_lists_by_comma and isinstance(cell, list):
                    cell = ", ".join(cell)

                bcc = render_conf.bool_conversion_config
                if bcc and isinstance(cell, bool):
                    cell = bcc.convert_true_to if cell else bcc.convert_false_to
                if isinstance(cell, TableCellLink):
                    cell = cell.to_html()
                elif render_conf.max_width and isinstance(cell, str):
                    cell = StringUtils.convert_string_to_multiline(
                        cell, max_line_length=render_conf.max_width, separator=render_conf.max_width_separator
                    )
                converted_row.append(cell)

            if render_conf.colorize_config:
                ResultPrinter._colorize_row(render_conf.colorize_config, converted_row, row)
            result.append(converted_row)

        return ConversionResult(src_data, result)

    @staticmethod
    def _colorize_row(conf: ColorizeConfig, converted_row, row):
        largest_range_end = -1
        for desc in conf.descriptors:
            range_end = desc.colorize_range[1]
            largest_range_end = max(largest_range_end, range_end)
        if largest_range_end >= len(converted_row):
            raise ValueError(
                "Invalid colorize config. Range is larger than number of columns in row. "
                "Colorize config: {}, row: {}".format(conf, row)
            )

        row_as_list = list(row)
        truthy = []
        for cd in conf.descriptors:
            filtered_type_values = list(
                filter(lambda x: type(x) == cd.type, row_as_list[cd.scan_range[0] : cd.scan_range[1]])
            )
            match_count = 0
            for idx, val in enumerate(filtered_type_values):
                if val == cd.value:
                    match_count += 1
            if cd.match_type == MatchType.ANY and match_count > 0:
                truthy.append(cd)
            elif cd.match_type == MatchType.ALL and match_count == len(filtered_type_values):
                truthy.append(cd)

        for cd in truthy:
            color_args = {cd.color_type.value: cd.color.value}
            for i in range(*cd.colorize_range):
                # Color multiline strings line by line
                if "\n" in str(converted_row[i]):
                    lines = converted_row[i].splitlines()
                    colored = []
                    for idx, line in enumerate(lines):
                        colored.append(color(line, **color_args))
                    converted_row[i] = "\n".join(colored)
                else:
                    converted_row[i] = color(converted_row[i], **color_args)
            if conf.eval_method == EvaluationMethod.FIRST_TRUTHY:
                break

    # TODO Users of this function should be migrated to _colorize_row and the new ResultPrinter
    @staticmethod
    def colorize_row(
        curr_row,
        convert_bools=False,
        char_if_present="X",
        char_if_not_present="-",
        color_if_okay="green",
        color_if_not_okay="red",
    ):
        res = []
        missing_backport = False
        if not all(curr_row[1:]):
            missing_backport = True

        # Mark first cell with red if any of the backports are missing
        # Mark first cell with green if all backports are present
        # Mark any bool cell with green if True, red if False
        for idx, cell in enumerate(curr_row):
            if (isinstance(cell, bool) and cell) or not missing_backport:
                if convert_bools and isinstance(cell, bool):
                    cell = char_if_present if cell else char_if_not_present
                res.append(color(cell, fore=color_if_okay))
            else:
                if convert_bools and isinstance(cell, bool):
                    cell = char_if_present if cell else char_if_not_present
                res.append(color(cell, fore=color_if_not_okay))
        return res


class BasicResultPrinter:
    @staticmethod
    def print_table(data, headers):
        LOG.info("Printing result table with format: fancy_grid")
        LOG.info(tabulate(data, headers, tablefmt=TabulateTableFormat.GRID.name))

    @staticmethod
    def print_table_html(data, headers):
        LOG.info("Printing result table with format: html")
        LOG.info(tabulate(data, headers, tablefmt=TabulateTableFormat.HTML.name))



import logging
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Any, List

import requests

from pythoncommons.os_utils import OsUtils

LOG = logging.getLogger(__name__)
GITHUB_PULLS_API_TEMPLATE = "https://api.github.com/repos/$owner$/$repo$/pulls/$PR_ID"
GITHUB_PULLS_LIST_API_TEMPLATE = "https://api.github.com/repos/$owner$/$repo$/pulls"
GITHUB_PULLS_LIST_API_QUERY_PAGE = "page"
GITHUB_PULLS_LIST_API_QUERY_PER_PAGE = "per_page"


class GitHubListPRsField(Enum):
    MERGEABLE = "mergeable"
    BASE = "base"
    HEAD = "head"
    NUMBER = "number"


class GitHubPRSrcDestField(Enum):
    REF = "ref"


@dataclass(frozen=True, eq=True)
class GitHubRepoIdentifier:
    owner: str
    repo: str

    def as_list_api(self):
        replacers = [("$owner$", self.owner), ("$repo$", self.repo)]
        return self._multi_replace(GITHUB_PULLS_LIST_API_TEMPLATE, replacers)

    def as_pulls_api(self):
        replacers = [("$owner$", self.owner), ("$repo$", self.repo)]
        return self._multi_replace(GITHUB_PULLS_API_TEMPLATE, replacers)

    @staticmethod
    def _multi_replace(base_str, replacers):
        res = base_str
        for replacer in replacers:
            res = res.replace(replacer[0], replacer[1])
        return res


class GithubPRMergeStatus(Enum):
    MERGEABLE = "Mergeable"
    NOT_MERGEABLE = "Not mergeable"
    UNKNOWN = "Unknown"
    PR_NOT_FOUND = "Pull request not found"


class GithubActionsEnvVar(Enum):
    CI_EXECUTION = "CI"
    GITHUB_ACTIONS = "GITHUB_ACTIONS"
    GITHUB_WORKSPACE = "GITHUB_WORKSPACE"


class GitHubUtils:
    ALL_PRS_BY_REPO_ID_CACHE: Dict[GitHubRepoIdentifier, Dict[Any, Any]] = {}

    @staticmethod
    def is_github_ci_execution() -> bool:
        is_github_ci_exec = OsUtils.get_env_value(GithubActionsEnvVar.GITHUB_ACTIONS.value)
        if is_github_ci_exec:
            LOG.debug("Detected Github Actions CI execution")
            return True
        return False

    @staticmethod
    def get_workspace_path() -> str:
        github_ws_path = OsUtils.get_env_value(GithubActionsEnvVar.GITHUB_WORKSPACE.value)
        if github_ws_path:
            LOG.debug("Detected Github Actions CI workspace path: %s", github_ws_path)
        return github_ws_path

    @staticmethod
    def is_pull_request_of_jira_mergeable(
        gh_repo_id: GitHubRepoIdentifier, jira_id: str, branches: List[str], use_cache=True
    ) -> Dict[str, GithubPRMergeStatus]:
        found_prs = GitHubUtils.find_pull_requests(gh_repo_id, jira_id, use_cache=use_cache)
        if not found_prs:
            return {br: GithubPRMergeStatus.PR_NOT_FOUND for br in branches}

        d = {}
        branches_set = set(branches)
        for pr in found_prs:
            pr_id = int(pr[GitHubListPRsField.NUMBER.value])
            target_branch_of_pr = pr[GitHubListPRsField.BASE.value][GitHubPRSrcDestField.REF.value]
            mergeable = GitHubUtils._is_pull_request_mergeable(gh_repo_id, pr_id)
            d[target_branch_of_pr] = mergeable
            branches_set.remove(target_branch_of_pr)

        # Loop through remaining branches and add PR_NOT_FOUND
        for br in branches_set:
            d[br] = GithubPRMergeStatus.PR_NOT_FOUND
        return d

    @staticmethod
    def _is_pull_request_mergeable(gh_repo_id: GitHubRepoIdentifier, pr_id: int) -> GithubPRMergeStatus:
        pr_url = GitHubUtils._get_pull_request_url(gh_repo_id, pr_id)
        pr_json = requests.get(pr_url).json()
        if GitHubListPRsField.MERGEABLE.value in pr_json:
            if pr_json[GitHubListPRsField.MERGEABLE.value]:
                return GithubPRMergeStatus.MERGEABLE
            else:
                return GithubPRMergeStatus.NOT_MERGEABLE
        return GithubPRMergeStatus.UNKNOWN

    @staticmethod
    def _get_pull_request_url(gh_repo_id: GitHubRepoIdentifier, pr_id: int):
        return gh_repo_id.as_pulls_api().replace("$PR_ID", str(pr_id))

    @classmethod
    def find_pull_requests(cls, gh_repo_id: GitHubRepoIdentifier, jira_id: str, use_cache=True) -> List[Any]:
        """
        With GitHub's pulls API, we can't get the number of PR results or the number of pages.
        However, we can get the PRs by specifying the page=<number> query parameter.
        We should query the PRs as long as the current page gave some meaningful result.
        If the resulted PR list is empty, we can stop the itaration.
        See:
        https://docs.github.com/en/rest/reference/pulls
        https://stackoverflow.com/a/38699904/1106893
        :param jira_id:
        :return:
        """
        # TODO This does not find PRs that are closed / merged
        all_pr_by_title = cls._find_all_pull_requests(gh_repo_id, use_cache=use_cache)
        return cls.find_in_all_pull_requests(all_pr_by_title, jira_id)

    @classmethod
    def find_in_all_pull_requests(cls, all_prs_by_title, jira_id) -> List[Any]:
        found_prs = []
        for title, pr_dict in all_prs_by_title.items():
            if title.startswith(jira_id):
                found_prs.append(pr_dict)
        return found_prs

    @classmethod
    def _find_all_pull_requests(cls, gh_repo_id: GitHubRepoIdentifier, use_cache=True):
        if use_cache and gh_repo_id in cls.ALL_PRS_BY_REPO_ID_CACHE:
            return cls.ALL_PRS_BY_REPO_ID_CACHE[gh_repo_id]

        all_prs_by_title = {}
        page_number = 1
        while True:
            page_param = f"{GITHUB_PULLS_LIST_API_QUERY_PAGE}={page_number}"
            url = f"{gh_repo_id.as_list_api()}?{GITHUB_PULLS_LIST_API_QUERY_PER_PAGE}=100&{page_param}"
            LOG.info("Querying Pull requests from URL: %s", url)
            resp = requests.get(url)
            if resp.status_code != 200:
                resp_json = resp.json()
                message = resp_json["message"]
                raise ValueError("HTTP error during querying {}. Message: {}".format(url, message))
            prs = resp.json()
            pr_by_title = {pr["title"]: pr for pr in prs}
            if pr_by_title:
                LOG.info(
                    "Found %d open PRs for URL: %s. All PRs found so far: %d",
                    len(pr_by_title),
                    url,
                    len(all_prs_by_title),
                )
            else:
                break
            all_prs_by_title.update(pr_by_title)
            page_number += 1
        LOG.info("Found %d open PRs for base URL: %s", len(all_prs_by_title), gh_repo_id.as_list_api())
        LOG.debug("Found open PRs for base URL '%s', details: %s", gh_repo_id.as_list_api(), all_prs_by_title)
        if use_cache:
            cls.ALL_PRS_BY_REPO_ID_CACHE[gh_repo_id] = all_prs_by_title
        return all_prs_by_title



import logging
import os
import platform
import sys
from abc import ABC, abstractmethod
from enum import Enum
from os.path import expanduser
import inspect
from typing import Dict, List

from pythoncommons.constants import PROJECT_NAME
from pythoncommons.date_utils import DateUtils
from pythoncommons.file_utils import FileUtils, FindResultType
from pythoncommons.os_utils import OsUtils
from pythoncommons.string_utils import StringUtils

LOG = logging.getLogger(__name__)


class ProjectUtilsEnvVar(Enum):
    OVERRIDE_USER_HOME_DIR = "OVERRIDE_USER_HOME_DIR"
    PROJECT_DETERMINATION_STRATEGY = "PYTHONCOMMONS_PROJECTUTILS_PROJECT_DETERMINATION_STRATEGY"


def determine_project_basedir():
    user_home = expanduser("~")
    override_user_home = OsUtils.get_env_value(ProjectUtilsEnvVar.OVERRIDE_USER_HOME_DIR.value)
    if override_user_home:
        LOG.info("Overriding user home dir with: %s", override_user_home)
        user_home = override_user_home
    return FileUtils.join_path(user_home, PROJECTS_BASEDIR_NAME)


MAC_PRIVATE_DIR = "private"
PROJECTS_BASEDIR_NAME = "snemeth-dev-projects"
PROJECTS_BASEDIR = determine_project_basedir()
REPOS_DIR = FileUtils.join_path(expanduser("~"), "development", "my-repos")
LOGS_DIR_NAME = "logs"
TEST_OUTPUT_DIR_NAME = "test"
TEST_LOG_FILE_POSTFIX = "TEST"
SITE_PACKAGES_DIRNAME = "site-packages"


class ProjectRootDeterminationStrategy(Enum):
    COMMON_FILE = "common_file"
    SYS_PATH = "sys_path"
    REPOSITORY_DIR = "repository_dir"


def get_sys_path_human_readable():
    return "\n".join(sys.path)


def get_stack_human_readable(stack):
    return "\n".join([str(f.frame) for f in stack])


class SimpleProjectUtils:
    @classmethod
    def get_project_dir(
        cls,
        basedir: str,
        parent_dir: str,
        dir_to_find: str,
        find_result_type: FindResultType,
        exclude_dirs: List[str] = None,
    ):
        found_dirs = FileUtils.find_files(
            basedir,
            find_type=find_result_type,
            regex=dir_to_find,
            parent_dir=parent_dir,
            exclude_dirs=exclude_dirs,
            single_level=False,
            full_path_result=True,
        )
        if len(found_dirs) != 1:
            raise ValueError(
                f"Expected to find 1 dir with name {dir_to_find} "
                f"and parent dir '{parent_dir}'. "
                f"Actual results: {found_dirs}"
            )
        return found_dirs[0]

    @classmethod
    def get_project_file(
        cls,
        basedir: str,
        file_to_find: str,
        find_result_type: FindResultType,
    ):
        found_files = FileUtils.find_files(
            basedir,
            find_type=find_result_type,
            regex=file_to_find,
            single_level=False,
            full_path_result=True,
        )
        if len(found_files) != 1:
            raise ValueError(f"Expected to find 1 file with name {file_to_find}." f"Actual results: {found_files}")
        return found_files[0]


class StrategyBase(ABC):
    @abstractmethod
    def determine_path(self, caller_file, stack, project_name_hint=None):
        pass

    def find_common_paths(self, path, file_of_caller):
        # Remove potential multiple slashes to check against the same path, see: https://stackoverflow.com/a/64459248/1106893
        file_of_caller = os.path.abspath(file_of_caller)
        found_match: bool = file_of_caller.startswith(path)
        if found_match:
            return path, file_of_caller

        found_match, new_file_of_caller = self.mac_specific_path_startswith(path, file_of_caller)
        LOG.debug("File of caller: '%s', path: '%s', found match: %s", file_of_caller, path, found_match)
        if found_match:
            return path, new_file_of_caller

        return None, file_of_caller

    @staticmethod
    def mac_specific_path_startswith(path, file_of_caller):
        # Remove potential multiple slashes to check against the same path, see: https://stackoverflow.com/a/64459248/1106893
        file_of_caller = os.path.abspath(file_of_caller)

        # Had to make an OS-based distinction here...
        # On MacOS, if the file is executed from /var or /tmp, the stackframe will contain the normal path.
        # However, even if the normal path is added to sys.path from a testcase,
        # it will be prepended with /private.
        # More on /private can be found here: https://apple.stackexchange.com/a/227869

        # Example scenario:
        # path: '/private/var/folders/nn/mkv5bwbd2fg8v8ztz5swpq980000gn/T/tmpp3k75qk2/python'
        # file_of_caller: '/var/folders/nn/mkv5bwbd2fg8v8ztz5swpq980000gn/T/tmpp3k75qk2/python/hello_world.py'
        is_mac = platform.system() == "Darwin"
        LOG.debug("Trying to match file '%s' with path '%s' with Mac-specific startswith.", file_of_caller, path)
        if is_mac and StringUtils.is_path_starting_with_dirname(path, MAC_PRIVATE_DIR):
            # WARNING: Cannot use os.path.join, neither StringUtils.prepend_path here as it removes /private from the path string :(
            # extended_file_of_caller = StringUtils.prepend_path(file_of_caller, MAC_PRIVATE_DIR)
            extended_file_of_caller = os.sep + MAC_PRIVATE_DIR + file_of_caller
            LOG.debug(
                f"Original file of caller: {file_of_caller}"
                f"\nExtended file of caller: {extended_file_of_caller}"
                f"\nPath: {path}"
            )
            if extended_file_of_caller.startswith(path):
                return True, extended_file_of_caller
        return False, file_of_caller

    LOG.debug(
        "Execution environment is not local, "
        "trying to determine project name with sys.path strategy. "
        f"Current sys.path: \n{get_sys_path_human_readable()}"
    )


class CommonPathStrategy(StrategyBase):
    def determine_path(self, file_of_caller, stack, project_name_hint=None):
        LOG.debug(
            "Execution environment is not local, "
            "trying to determine project name with common files strategy. "
            f"Current sys.path: \n{get_sys_path_human_readable()}"
            f"Current caller file: {file_of_caller}"
        )
        project_root_path, visited_paths = FileUtils.find_repo_root_dir_auto(file_of_caller, raise_error=False)
        if project_root_path == os.sep:
            orig_path = os.path.realpath(file_of_caller)
            err_message = (
                f"Failed to find project root directory starting from path '{orig_path}'. "
                f"Visited: {visited_paths}\n"
                f"Strategy: {type(self).__name__}\n"
                f"Caller file: {file_of_caller}"
            )
            if project_name_hint:
                LOG.error(err_message + " Returning project name from hint: " + project_name_hint)
                return file_of_caller, project_name_hint
            else:
                raise ValueError(err_message)

        LOG.debug(f"Found project root: {project_root_path}")
        comps = FileUtils.get_path_components(project_root_path)
        project = comps[-1]
        path = comps[0:-1]
        LOG.info(f"Determined path: {path}, project: {project}")
        return path, project


class RepositoryDirStrategy(StrategyBase):
    def determine_path(self, file_of_caller, stack, project_name_hint=None):
        LOG.debug(
            "Trying to determine project name with repository dir strategy. "
            f"Current sys.path: \n{get_sys_path_human_readable()}"
        )
        filename = file_of_caller[len(REPOS_DIR) :]
        # We should return the first dir name of the path
        # Cut leading slashes, if any as split would return empty string for 0th component
        filename = StringUtils.strip_leading_os_sep(filename)
        project = StringUtils.get_first_dir_of_path(filename)
        LOG.info(f"Determined path: {REPOS_DIR}, project: {project}")
        return REPOS_DIR, project


class SysPathStrategy(StrategyBase):
    def determine_path(self, file_of_caller, stack, project_name_hint=None):
        for path in sys.path:
            LOG.debug("Checking path: '%s' against file_of_caller: '%s'", path, file_of_caller)
            if ProjectUtils.FORCE_SITE_PACKAGES_IN_PATH_NAME and SITE_PACKAGES_DIRNAME not in path:
                LOG.debug("Skipping path: '%s', as '%s' not found in the path", path, SITE_PACKAGES_DIRNAME)
                continue
            matched_base_path, file_of_caller = self.find_common_paths(path, file_of_caller)
            if not matched_base_path:
                continue

            LOG.debug("Found Base path: %s for file of caller: %s", matched_base_path, file_of_caller)
            matched_base_path = StringUtils.strip_trailing_os_sep(matched_base_path)
            if ProjectUtils.FORCE_SITE_PACKAGES_IN_PATH_NAME and not matched_base_path.endswith(SITE_PACKAGES_DIRNAME):
                LOG.debug(
                    "Matched base path does not end with '%s'. Dropping path components after it.",
                    SITE_PACKAGES_DIRNAME,
                )
                # Need to cut the last dir from the path, so we find the site-packages root
                # Example: /<somepath>/venv/lib/python3.8/site-packages/test_project
                matched_base_path = StringUtils.remove_last_dir_from_path(matched_base_path)
                LOG.debug("Final base path: %s", matched_base_path)

            parts = file_of_caller.split(matched_base_path)
            LOG.debug(f"Parts of path after split: {parts}")

            # Example #1: ['', '/test_project.py'] --> Need to get item with index 1
            # Cut leading slashes
            proj_name = parts[1]
            proj_name = StringUtils.strip_leading_os_sep(proj_name)

            # Example #2: ['', '/testproject/commands/testcommand/dummy_test_command.py']
            if StringUtils.is_path_multi_component(proj_name):
                LOG.debug(
                    "Found multiple dirs in project name: %s. " "Assuming first dir is the name of the project.",
                    proj_name,
                )
                proj_name = StringUtils.get_first_dir_of_path_if_multi_component(proj_name)
            LOG.info(f"Determined path: {matched_base_path}, project: {proj_name}")
            return matched_base_path, proj_name

        err_message = (
            f"Cannot determine project! "
            f"File of caller: {file_of_caller}\n"
            f"Call stack: \n{get_stack_human_readable(stack)}"
            f"Strategy: {type(self).__name__}"
        )
        if project_name_hint:
            LOG.error(err_message + " Returning project name from hint: " + project_name_hint)
            return file_of_caller, project_name_hint
        else:
            raise ValueError(err_message)


class ProjectUtils:
    PROJECT_BASEDIR_DICT = {}
    CHILD_DIR_DICT = {}
    CHILD_DIR_TEST_DICT = {}
    FILES_TO_PROJECT = {}
    FORBIDDEN_DIR_NAMES = ["unittest", "pytest", "_pytest"]
    test_execution: bool = False
    project_root_determine_strategy_set_manually: bool = False
    default_project_determine_strategy = ProjectRootDeterminationStrategy.COMMON_FILE
    project_root_determine_strategy = default_project_determine_strategy
    FORCE_SITE_PACKAGES_IN_PATH_NAME = True
    STRATEGIES: Dict[ProjectRootDeterminationStrategy, StrategyBase] = {
        ProjectRootDeterminationStrategy.COMMON_FILE: CommonPathStrategy(),
        ProjectRootDeterminationStrategy.SYS_PATH: SysPathStrategy(),
        ProjectRootDeterminationStrategy.REPOSITORY_DIR: RepositoryDirStrategy(),
    }

    @classmethod
    def reset_root_determine_strategy_to_default(cls):
        cls.project_root_determine_strategy_set_manually = False
        cls.project_root_determine_strategy = cls.default_project_determine_strategy

    @classmethod
    def set_root_determine_strategy(cls, strategy: ProjectRootDeterminationStrategy, allow_overwrite=True):
        old_strategy = cls.project_root_determine_strategy
        LOG.info("Discovered project root determine strategy: %s", old_strategy)
        if not cls.project_root_determine_strategy_set_manually or allow_overwrite:
            LOG.info(
                "Overwriting project root determine strategy! Old value: %s, New value: %s", old_strategy, strategy
            )
            cls.project_root_determine_strategy = strategy
            cls.project_root_determine_strategy_set_manually = True

    @classmethod
    def determine_project_and_parent_dir(cls, file_of_caller, stack, project_name_hint=None):
        received_args = locals().copy()
        received_args["stack"] = get_stack_human_readable(stack)
        LOG.debug(
            f"Determining project name with strategy '{cls.project_root_determine_strategy}'. "
            f"Received args: {received_args}. \n"
            f"{cls._get_known_projects_str()}\n"
        )

        if file_of_caller in cls.FILES_TO_PROJECT:
            project = cls.FILES_TO_PROJECT[file_of_caller]
            LOG.debug(f"Found cached project name '{project}', file was already a caller: {file_of_caller}")
            return file_of_caller, project

        strategy: StrategyBase = cls._determine_strategy(file_of_caller)
        path, project = strategy.determine_path(file_of_caller, stack, project_name_hint=project_name_hint)
        cls.FILES_TO_PROJECT[file_of_caller] = project
        return path, project

        # TODO Can this happen?
        # raise ValueError(
        #     f"Unexpected project execution directory. \n"
        #     f"Filename of caller: '{file_of_caller}'\n"
        #     f"Printing diagnostic info including call stack + sys.path...\n"
        #     f"\nCall stack: \n{get_stack_human_readable(stack)}\n"
        #     f"\nsys.path: \n{get_sys_path_human_readable()}")

    @classmethod
    def _determine_strategy(cls, file_of_caller) -> StrategyBase:
        strategy_env_var = OsUtils.get_env_value(ProjectUtilsEnvVar.PROJECT_DETERMINATION_STRATEGY.value)
        if strategy_env_var:
            strat = ProjectRootDeterminationStrategy[strategy_env_var.upper()]
            LOG.info("Using strategy '%s' based on env var", strat)
            return cls.STRATEGIES[strat]
        strategy: StrategyBase = cls.STRATEGIES[cls.project_root_determine_strategy]
        if (
            REPOS_DIR in file_of_caller
            and cls.project_root_determine_strategy == cls.default_project_determine_strategy
        ):
            strategy = cls.STRATEGIES[ProjectRootDeterminationStrategy.REPOSITORY_DIR]
        return strategy

    @classmethod
    def get_output_basedir(
        cls,
        basedir_name: str,
        ensure_created=True,
        allow_python_commons_as_project=False,
        basedir=PROJECTS_BASEDIR,
        project_name_hint=None,
    ):
        if not basedir_name:
            raise ValueError("Basedir name should be specified!")

        project_name = cls.verify_caller_filename_valid(
            allow_python_commons_as_project=allow_python_commons_as_project, project_name_hint=project_name_hint
        )
        proj_basedir = FileUtils.join_path(basedir, basedir_name)
        if project_name in cls.PROJECT_BASEDIR_DICT:
            old_basedir = cls.PROJECT_BASEDIR_DICT[project_name]
            if old_basedir != proj_basedir:
                raise ValueError(
                    "Project is already registered with a different output basedir. Details: \n"
                    f"Old basedir name: {StringUtils.get_last_dir_of_path(old_basedir)}\n"
                    f"Project basedir's old full path: {old_basedir}\n"
                    f"New basedir name would be: {basedir_name}\n"
                    f"Project basedir's new full path would be: {proj_basedir}\n"
                )
        cls.PROJECT_BASEDIR_DICT[project_name] = proj_basedir

        if ensure_created:
            FileUtils.ensure_dir_created(proj_basedir)
        return proj_basedir

    @classmethod
    def get_test_output_basedir(cls, basedir_name: str, allow_python_commons_as_project=False):
        """

        :param basedir_name:
        :param allow_python_commons_as_project: This is useful and a must for test executions of ProjectUtils (e.g. JiraUtilsTests)
        as stackframes calling pythoncommons are only the methods of the unittest framework.
        :return:
        """
        cls.test_execution = True
        project_name = cls.verify_caller_filename_valid(allow_python_commons_as_project=allow_python_commons_as_project)
        if project_name not in cls.PROJECT_BASEDIR_DICT:
            # Creating project dir for the first time
            proj_basedir = cls.get_output_basedir(
                basedir_name, allow_python_commons_as_project=allow_python_commons_as_project
            )
        else:
            proj_basedir = cls.PROJECT_BASEDIR_DICT[project_name]

        return FileUtils.join_path(proj_basedir, TEST_OUTPUT_DIR_NAME)

    @classmethod
    def get_output_child_dir(cls, dir_name: str, ensure_created=True, project_name_hint=None):
        if not dir_name:
            raise ValueError("Dir name should be specified!")
        project_name = cls._validate_project_for_child_dir_creation(project_name_hint=project_name_hint)

        if project_name in cls.CHILD_DIR_DICT and dir_name in cls.CHILD_DIR_DICT[project_name]:
            stored_dir = cls.CHILD_DIR_DICT[project_name][dir_name]
            LOG.debug(f"Found already stored child dir for project '{project_name}': {stored_dir}")
            FileUtils.ensure_dir_created(stored_dir)
            return stored_dir

        proj_basedir = cls.PROJECT_BASEDIR_DICT[project_name]
        new_child_dir = FileUtils.join_path(proj_basedir, dir_name)
        if project_name not in cls.CHILD_DIR_DICT:
            cls.CHILD_DIR_DICT[project_name] = {}
        cls.CHILD_DIR_DICT[project_name][dir_name] = new_child_dir

        if ensure_created:
            FileUtils.ensure_dir_created(new_child_dir)
        return new_child_dir

    @classmethod
    def get_test_output_child_dir(
        cls, dir_name: str, ensure_created=True, special_parent_dir=None, project_name_hint=None
    ):
        if not dir_name:
            raise ValueError("Dir name should be specified!")
        project_name = cls._validate_project_for_child_dir_creation(project_name_hint=project_name_hint)

        if project_name in cls.CHILD_DIR_TEST_DICT and dir_name in cls.CHILD_DIR_TEST_DICT[project_name]:
            stored_dir = cls.CHILD_DIR_TEST_DICT[project_name][dir_name]
            LOG.debug(f"Found already stored child test dir for project '{project_name}': {stored_dir}")
            FileUtils.ensure_dir_created(stored_dir)
            return stored_dir

        if special_parent_dir:
            if not FileUtils.does_path_exist(special_parent_dir):
                raise ValueError(f"Specified parent dir does not exist: {special_parent_dir}")
            LOG.debug(f"Parent dir of new child directory will be: {special_parent_dir}")
            parent_dir = special_parent_dir
            new_child_dir = FileUtils.join_path(parent_dir, dir_name)
        else:
            # Default parent dir: Basedir of project
            # New child dir: basedir/test/<new child dir name>
            parent_dir = cls.PROJECT_BASEDIR_DICT[project_name]
            new_child_dir = FileUtils.join_path(parent_dir, TEST_OUTPUT_DIR_NAME, dir_name)
        if project_name not in cls.CHILD_DIR_TEST_DICT:
            cls.CHILD_DIR_TEST_DICT[project_name] = {}
        cls.CHILD_DIR_TEST_DICT[project_name][dir_name] = new_child_dir

        if ensure_created:
            FileUtils.ensure_dir_created(new_child_dir)
        return new_child_dir

    @classmethod
    def get_session_dir_under_child_dir(cls, child_dir_name: str):
        # If this method called from production code but we are exeucting it from test code
        if cls.test_execution:
            return cls._get_session_dir_under_child_dir(child_dir_name, test=True)
        return cls._get_session_dir_under_child_dir(child_dir_name)

    @classmethod
    def get_test_session_dir_under_child_dir(cls, child_dir_name: str):
        return cls._get_session_dir_under_child_dir(child_dir_name, test=True)

    @classmethod
    def _get_session_dir_under_child_dir(cls, child_dir_name, test: bool = False):
        child_dir_type: str = "child dir" if not test else "test child dir"
        dir_dict = cls.CHILD_DIR_DICT if not test else cls.CHILD_DIR_TEST_DICT

        if not child_dir_name:
            raise ValueError(f"Project {child_dir_type} name should be specified!")

        project_name = cls._validate_project_for_child_dir_creation()
        if project_name in dir_dict and child_dir_name in dir_dict[project_name]:
            stored_dir = dir_dict[project_name][child_dir_name]
            LOG.debug(f"Found already stored {child_dir_type} for project '{project_name}': {stored_dir}")

            session_dir = FileUtils.join_path(stored_dir, f"session-{DateUtils.now_formatted('%Y%m%d_%H%M%S')}")
            FileUtils.ensure_dir_created(session_dir)
            return session_dir
        else:
            raise ValueError(
                f"Cannot find stored {child_dir_type} for project. "
                f"Project: {project_name}, "
                f"Child dir: {child_dir_name}, "
                f"All stored {child_dir_type}s: {dir_dict}"
            )

    @classmethod
    def save_to_test_file(cls, dir_name: str, filename: str, file_contents: str):
        if not dir_name:
            raise ValueError("Dir name should be specified!")
        if not filename:
            raise ValueError("Filename should be specified!")

        project_name = cls._validate_project_for_child_dir_creation()
        cls.validate_test_child_dir(dir_name, project_name)
        dir_path = cls.CHILD_DIR_TEST_DICT[project_name][dir_name]
        FileUtils.save_to_file(FileUtils.join_path(dir_path, filename), file_contents)

    @classmethod
    def validate_test_child_dir(cls, dir_name, project_name):
        if dir_name not in cls.CHILD_DIR_TEST_DICT[project_name]:
            raise ValueError(f"Unknown test child dir with name '{dir_name}' for project '{project_name}'.")

    @classmethod
    def remove_test_files_and_recreate_dir(cls, dir_name: str, clazz):
        project_name = cls._validate_project_for_child_dir_creation()
        cls.validate_test_child_dir(dir_name, project_name)
        dir_path = cls.CHILD_DIR_TEST_DICT[project_name][dir_name]

        LOG.info(f"Removing dir: {dir_path}")
        FileUtils.remove_files(dir_path, ".*")
        LOG.info(f"Recreating dir: {dir_path}")
        new_dir = FileUtils.ensure_dir_created(dir_path)
        LOG.info("Basedir of %s is: %s", clazz.__name__, new_dir)
        return new_dir

    @classmethod
    def _validate_project_for_child_dir_creation(cls, project_name_hint=None):
        project_name = cls.verify_caller_filename_valid(project_name_hint=project_name_hint)
        if project_name not in cls.PROJECT_BASEDIR_DICT:
            raise ValueError(
                f"Project '{project_name}' is unknown. "
                f"{cls._get_known_projects_str()}\n"
                f"Please call {ProjectUtils.__name__}.{ProjectUtils.get_output_basedir.__name__} "
                f"first in order to set the basedir for the project!"
            )
        return project_name

    @classmethod
    def get_default_log_file(cls, project_name: str, postfix: str = None, level_name: str = None):
        return cls._get_log_filename_internal(project_name, postfix, level_name=level_name, prod=True)

    @classmethod
    def get_default_test_log_file(cls, project_name: str, postfix: str = None, level_name: str = None):
        return cls._get_log_filename_internal(project_name, postfix, level_name=level_name, prod=False)

    @classmethod
    def _get_log_filename_internal(
        cls, project_name: str, postfix: str = None, level_name: str = None, prod: bool = True
    ):
        if postfix:
            postfix = "-" + postfix
        else:
            postfix = ""
        if not prod:
            postfix += "-" + TEST_LOG_FILE_POSTFIX

        if level_name:
            level_name = "-" + level_name
        else:
            level_name = ""

        filename = f"{project_name}{postfix}{level_name}-{DateUtils.get_current_datetime()}"
        log_dir = (
            cls.get_logs_dir(project_name_hint=project_name)
            if prod
            else cls.get_test_logs_dir(project_name_hint=project_name)
        )
        return FileUtils.join_path(log_dir, filename)

    @classmethod
    def get_logs_dir(cls, project_name_hint=None):
        return cls.get_output_child_dir(LOGS_DIR_NAME, project_name_hint=project_name_hint)

    @classmethod
    def get_test_logs_dir(cls, project_name_hint=None):
        return cls.get_test_output_child_dir(LOGS_DIR_NAME, project_name_hint=project_name_hint)

    @classmethod
    def verify_caller_filename_valid(cls, allow_python_commons_as_project=False, project_name_hint=None):
        stack = inspect.stack()
        stack_frame, idx = cls._find_first_non_pythoncommons_stackframe(stack)
        file_of_caller = stack_frame.filename
        LOG.debug("Filename of caller: " + file_of_caller)
        if StringUtils.is_any_of_dir_names_in_path(file_of_caller, cls.FORBIDDEN_DIR_NAMES):
            message = (
                f"Detected caller as 'unittest'. Current stack frame: {stack_frame}\n"
                f"Stack: {get_stack_human_readable(stack)}"
            )
            if allow_python_commons_as_project:
                LOG.warning(message)
                # Get the previous frame which should belong to pythoncommons
                python_commons_frame = stack[idx - 1]
                file_of_caller = python_commons_frame.filename
            else:
                message += (
                    "\n'allow_python_commons_as_project' is set to False. "
                    "Please set 'allow_python_commons_as_project' to True "
                    "to the ProjectUtils method that initiated the call."
                )
                raise ValueError(message)
        path, project = cls.determine_project_and_parent_dir(file_of_caller, stack, project_name_hint=project_name_hint)
        return project

    @classmethod
    def _find_first_non_pythoncommons_stackframe(cls, stack):
        idx = 1
        while idx < len(stack):
            LOG.debug("Inspecting stack frame: %s", stack[idx])
            fname = stack[idx].filename
            if PROJECT_NAME not in fname:
                break
            idx += 1
        if idx == len(stack):
            # Walked up the stack and haven't found any frame that is not pythoncommons
            raise ValueError(
                "Walked up the stack but haven't found any frame that does not belong to python-commons. \n"
                "Printing the stack: \n"
                f"{get_stack_human_readable(stack)}"
            )
        return stack[idx], idx

    @classmethod
    def _get_known_projects_str(cls):
        return f"Known projects are: {list(cls.PROJECT_BASEDIR_DICT.keys())}"

    @classmethod
    def get_all_project_basedirs(cls):
        return cls.PROJECT_BASEDIR_DICT.values()

    @classmethod
    def get_project_basedirs_dict(cls):
        return cls.PROJECT_BASEDIR_DICT



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pythoncommons',
 'pythoncommons.file_parser',
 'pythoncommons.test-scripts',
 'pythoncommons.test-scripts.commands.testcommand',
 'pythoncommons.tests']

package_data = \
{'': ['*']}

install_requires = \
['Colr>=0.9.1,<0.10.0',
 'GitPython>=3.1.27,<4.0.0',
 'bs4>=0.0.1,<0.0.2',
 'dataclasses-json>=0.5.7,<0.6.0',
 'docker>=6.0.0,<7.0.0',
 'google-api-python-client==1.7.9',
 'google-auth-httplib2==0.0.3',
 'google-auth-oauthlib==0.4.1',
 'gspread>=5.5.0,<6.0.0',
 'humanize>=4.4.0,<5.0.0',
 'jira>=3.4.1,<4.0.0',
 'oauth2client>=4.1.3,<5.0.0',
 'pytest>=6.2.3,<6.3.0',
 'requests>=2.28.1,<3.0.0',
 'sh>=1.14.1,<1.15.0',
 'tabulate>=0.8.10,<0.9.0']

setup_kwargs = {
    'name': 'python-common-lib',
    'version': '1.0.1',
    'description': '',
    'long_description': '# python-commons\n\nRun ./setup.sh to set up git pre/post push hook scripts.\nThen, a similar script loaded to the environment will execute the pre/post push hook scripts: \nhttps://stackoverflow.com/a/3812238/1106893\n\nFor example loading this script and defining an alias like this will do the trick:\n`alias gpwh="git-push-with-hooks.sh"`\n\n\n## Setup of precommit\n\nConfigure precommit as described in this blogpost: https://ljvmiranda921.github.io/notebook/2018/06/21/precommits-using-black-and-flake8/\nCommands:\n1. Install precommit: `pip install pre-commit`\n2. Make sure to add pre-commit to your path. For example, on a Mac system, pre-commit is installed here: \n   `$HOME/Library/Python/3.8/bin/pre-commit`.\n2. Execute `pre-commit install` to install git hooks in your `.git/` directory.\n\n## Troubleshooting\n\n### Installation issues\nIn case you\'re facing a similar issue:\n```\nAn error has occurred: InvalidManifestError: \n=====> /<userhome>/.cache/pre-commit/repoBP08UH/.pre-commit-hooks.yaml does not exist\nCheck the log at /<userhome>/.cache/pre-commit/pre-commit.log\n```\n, please run: `pre-commit autoupdate`\nMore info here: https://github.com/pre-commit/pre-commit/issues/577',
    'author': 'Szilard Nemeth',
    'author_email': 'szilard.nemeth88@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
import math
import os
import string
import re
import unicodedata
from enum import Enum
from typing import List

import pythoncommons.file_utils as fileutils

LOG = logging.getLogger(__name__)


def auto_str(cls, with_repr=True):
    def __str__(self):
        return "%s(%s)" % (type(self).__name__, ", ".join("%s=%s" % item for item in vars(self).items()))

    cls.__str__ = __str__

    def __repr__(self):
        return __str__(self)

    if with_repr:
        cls.__repr__ = __repr__
    return cls


# TODO complete this implementation
def auto_str2(cls, with_repr=True, exclude_props=None):
    if not exclude_props:
        exclude_props = []

    def __str__(self):
        props = vars(self).items()
        filtered_props = filter(lambda x: x not in exclude_props, props)
        ret = "%s(%s)" % (type(self).__name__, ", ".join("%s=%s" % prop for prop in filtered_props))
        return ret

    cls.__str__ = __str__

    def __repr__(self):
        return __str__(self)

    if with_repr:
        cls.__repr__ = __repr__
    return cls


def auto_repr(cls):
    def __repr__(self):
        return self.__str__()

    cls.__repr__ = __repr__
    return cls


class StringUtils:
    class StripMode(Enum):
        BEGINNING = 0
        END = 1

    @staticmethod
    def list_to_multiline_string(list):
        return "\n".join(str(x) for x in list)

    @staticmethod
    def dict_to_multiline_string(dict):
        return "\n".join([f"{k}: {v}" for k, v in dict.items()])

    @staticmethod
    def make_piped_string(strings: List[str]):
        return "|".join(strings)

    @staticmethod
    def get_first_line_of_multiline_str(multi_line_str):
        return StringUtils.get_line_of_multi_line_str(multi_line_str, 0)

    @staticmethod
    def get_line_of_multi_line_str(multi_line_str, line_number):
        if "\n" not in multi_line_str:
            raise ValueError("String is not a multi line string.")
        return multi_line_str.split("\n")[line_number]

    @staticmethod
    def replace_last(s, to_replace, replace_with, count):
        split = s.rsplit(to_replace, count)
        return replace_with.join(split)

    @staticmethod
    def replace_special_chars(unistr):
        if not isinstance(unistr, str):
            LOG.warning("Object expected to be unicode: " + str(unistr))
            return str(unistr)
        normalized = unicodedata.normalize("NFD", unistr).encode("ascii", "ignore")
        normalized = normalized.decode("utf-8")
        valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
        valid_title = "".join(c for c in normalized if c in valid_chars)
        return valid_title

    @staticmethod
    def convert_string_to_multiline(string, max_line_length, separator=" "):
        if not len(string) > max_line_length:
            return string

        result = ""
        curr_line_length = 0
        parts = string.split(separator)
        for idx, part in enumerate(parts):
            if curr_line_length + len(part) < max_line_length:
                result += part
                # Add length of part + 1 for space to current line length, if required
                curr_line_length += len(part)
            else:
                result += "\n"
                result += part
                curr_line_length = len(part)

            # If not last one, add separator
            if not idx == len(parts) - 1:
                result += separator
                curr_line_length += 1
        return result

    @staticmethod
    def count_leading_zeros(s):
        count = 0
        for i in range(len(s)):
            if s[i] != "0":
                return count
            count += 1

    @staticmethod
    def increase_numerical_str(string):
        num_zeros = StringUtils.count_leading_zeros(string)
        format_str = "%0" + str(num_zeros + 1) + "d"
        return format_str % (int(string) + 1)

    @staticmethod
    def generate_header_line(string, char="=", length=80):
        result = ""
        fill_length = length - len(string)
        filler = char * (math.floor(fill_length / 2))
        result += filler
        result += string
        result += filler
        return result

    @staticmethod
    def strip_strings(filename, strip_strs):
        for str, mode in strip_strs:
            if not isinstance(mode, StringUtils.StripMode):
                raise ValueError(
                    f"Mode of operation is invalid. " f"It should be an instance of: {StringUtils.StripMode.__name__}"
                )
            if mode == StringUtils.StripMode.BEGINNING:
                if filename.startswith(str):
                    filename = filename.split(str)[1]
            elif mode == StringUtils.StripMode.END:
                if filename.endswith(str):
                    filename = filename.split(str)[0]
        LOG.debug("Stripped string: " + filename)
        return filename

    @staticmethod
    def escape_strs(strings: List[str]):
        escaped_lines = []
        for orig_str in strings:
            mod_str = StringUtils.escape_str(orig_str)
            if orig_str != mod_str:
                LOG.debug(f"Modified line. Original: {orig_str}, Modified: {mod_str}")
            escaped_lines.append(orig_str)
        return escaped_lines

    @staticmethod
    def escape_str(orig_str, escape_single_quotes=True, escape_double_quotes=True):
        mod_str = orig_str
        if escape_single_quotes:
            mod_str = mod_str.replace("'", "\\'")
        if escape_double_quotes:
            mod_str = mod_str.replace('"', '\\"')
        return mod_str

    @staticmethod
    def wrap_to_quotes(s):
        return f'"{s}"'

    @staticmethod
    def wrap_to_single_quotes(s):
        return f"'{s}'"

    @staticmethod
    def strip_leading_os_sep(path):
        if path.startswith(os.sep):
            return path[1:]
        return path

    @staticmethod
    def strip_trailing_os_sep(path):
        if path.endswith(os.sep):
            return path[:-1]
        return path

    @staticmethod
    def get_first_dir_of_path(path):
        components = path.split(os.sep)
        if not components or len(components) == 0:
            raise ValueError(f"Invalid path: {path}, can't extract first directory!")
        return components[0]

    @staticmethod
    def get_last_dir_of_path(path):
        components = path.split(os.sep)
        if not components or len(components) == 0:
            raise ValueError(f"Invalid path: {path}, can't extract first directory!")
        return components[-1]

    @staticmethod
    def remove_last_dir_from_path(path):
        path_components = path.split(os.sep)
        path_components = path_components[:-1]
        return os.sep.join(path_components)

    @staticmethod
    def is_path_starting_with_dirname(path, dir_name):
        return path.startswith(os.sep + dir_name)

    @staticmethod
    def prepend_path(orig_path, prepend_with):
        return fileutils.FileUtils.join_path(os.sep, prepend_with, orig_path)

    @staticmethod
    def get_first_dir_of_path_if_multi_component(path):
        if os.sep in path:
            return path.split(os.sep)[0]
        return path

    @staticmethod
    def is_path_multi_component(path):
        if os.sep in path:
            return True
        return False

    @staticmethod
    def is_dir_name_in_path(path, dir_name):
        parts = path.split(os.sep)
        if dir_name in parts:
            return True
        return False

    @staticmethod
    def is_any_of_dir_names_in_path(path, dir_names: List[str]):
        return any(StringUtils.is_dir_name_in_path(path, dirname) for dirname in dir_names)

    @staticmethod
    def get_list_of_components_from_path(path):
        if os.sep in path:
            return path.split(os.sep)
        else:
            return [path]

    @staticmethod
    def find_all(s, expr):
        start = 0
        while True:
            start = s.find(expr, start)
            if start == -1:
                return
            yield start
            start += len(expr)  # use start += 1 to find overlapping matches


class RegexUtils:
    @staticmethod
    def filter_list_by_regex(list, regex):
        p = re.compile(regex)
        return [s for s in list if p.match(s)]

    @staticmethod
    def ensure_matches_pattern(string, regex, raise_exception=False):
        regex_obj = re.compile(regex)
        result = regex_obj.match(string)
        if raise_exception and not result:
            raise ValueError("String '{}' does not match regex pattern: {}".format(string, regex))
        return result

    @staticmethod
    def get_matched_group(str, regex, group):
        match = re.match(regex, str)
        if not match or len(match.groups()) < group:
            raise ValueError(
                "String '{}' does not have match with group number '{}'. Regex: '{}', Match object: '{}'",
                str,
                group,
                regex,
                match,
            )
        return match.group(group)



import logging
import os
import sys

from pythoncommons.project_utils import ProjectUtils

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.DEBUG)


class FakeBranchComparator:
    def __init__(self):
        output_child_dir = ProjectUtils.get_output_child_dir("branch-comparator")
        session_dir = ProjectUtils.get_session_dir_under_child_dir(os.path.basename(output_child_dir))
        LOG.info("Session dir: " + session_dir)



import logging
import os
import sys
from dataclasses import dataclass, field
from enum import Enum
from logging.handlers import TimedRotatingFileHandler
import logging.config
from typing import List, Dict, Callable

from _pytest.logging import _LiveLoggingStreamHandler
from _pytest.terminal import TerminalReporter
from pythoncommons.constants import PROJECT_NAME as PYTHONCOMMONS_PROJECT_NAME, ExecutionMode
from pythoncommons.date_utils import DateUtils
from pythoncommons.file_utils import FileUtils
from pythoncommons.project_utils import ProjectUtils
from pythoncommons.test_utils import PyTestUtils

DEFAULT_CONSOLE_STREAM = sys.stdout
DEFAULT_LOG_YAML_FILENAME = "logging_default.yaml"
DEFAULT_FORMAT = "%(asctime)s - %(levelname)s - %(name)s - %(message)s"


def _get_timestamp_as_str():
    return DateUtils.now_formatted("%Y_%m_%d_%H_%M_%S")


def get_project_logger(project_name, module_name, postfix=None):
    if postfix:
        if postfix.startswith("."):
            postfix = postfix[1:]
        return logging.getLogger("{}.{}.{}".format(project_name, module_name, postfix))
    return logging.getLogger("{}.{}".format(project_name, module_name))


def get_root_logger():
    return logging.getLogger("")


class HandlerType(Enum):
    FILE = "file"
    CONSOLE = "console"


_ROOT_LOG = get_root_logger()
DEFAULT_LOG_LEVEL: int = logging.INFO


@dataclass
class SimpleLoggingSetupConfig:
    project_name: str
    execution_mode: ExecutionMode
    file_log_level_name: str
    console_log_level_name: str
    formatter: str
    console_handler: logging.StreamHandler
    file_handlers: List[logging.FileHandler]
    main_project_logger: logging.Logger
    log_file_paths: Dict[int, str]


@dataclass
class SimpleLoggingSetupInputConfig:
    project_name: str
    logger_name_prefix: str
    debug: bool = False
    console_debug: bool = False
    format_str: str = None
    file_postfix: str = None
    execution_mode: ExecutionMode = ExecutionMode.PRODUCTION
    modify_pythoncommons_logger_names: bool = True
    remove_existing_handlers: bool = True
    disable_propagation: bool = True
    enable_logging_setup_debug_details: bool = False

    # Dynamic fields
    specified_file_log_level: int or None = None
    handlers: List[logging.Handler] = field(default_factory=list)
    project_main_logger: logging.Logger or None = None


# TODO this is copied from another project, eliminate duplication later
class SimpleLoggingSetup:
    _ALL_LOG_FILES = []

    @staticmethod
    def init_logger(
        project_name: str,
        logger_name_prefix: str,
        execution_mode: ExecutionMode,
        console_debug=False,
        postfix: str = None,
        repos=None,
        verbose_git_log=False,
        format_str=None,
        sanity_check_number_of_handlers=True,
        enable_logging_setup_debug_details: bool = False,
    ) -> SimpleLoggingSetupConfig:
        if not project_name:
            raise ValueError("Project name must be specified!")
        if not logger_name_prefix:
            raise ValueError("Logger name prefix must be specified!")

        final_format_str = DEFAULT_FORMAT
        if format_str:
            final_format_str = format_str
        logging_config: SimpleLoggingSetupConfig = SimpleLoggingSetup.init_logging(
            project_name=project_name,
            logger_name_prefix=logger_name_prefix,
            debug=True,
            console_debug=console_debug,
            format_str=final_format_str,
            file_postfix=postfix,
            execution_mode=execution_mode,
            sanity_check_number_of_handlers=sanity_check_number_of_handlers,
            enable_logging_setup_debug_details=enable_logging_setup_debug_details,
        )
        SimpleLoggingSetup._setup_gitpython_log(repos, verbose_git_log)
        return logging_config

    @staticmethod
    def _setup_gitpython_log(repos, verbose_git_log):
        # https://gitpython.readthedocs.io/en/stable/tutorial.html#git-command-debugging-and-customization
        # THIS WON'T WORK BECAUSE GITPYTHON MODULE IS LOADED BEFORE THIS CALL
        # os.environ["GIT_PYTHON_TRACE"] = "1"
        # https://github.com/gitpython-developers/GitPython/issues/222#issuecomment-68597780
        logging.getLogger().warning("Cannot enable GIT_PYTHON_TRACE because repos list is empty!")
        if repos:
            for repo in repos:
                val = "full" if verbose_git_log else "1"
                type(repo.git).GIT_PYTHON_TRACE = val

    @staticmethod
    def init_logging(
        project_name: str,
        logger_name_prefix: str,
        debug: bool = False,
        console_debug: bool = False,
        format_str: str = None,
        file_postfix: str = None,
        execution_mode: ExecutionMode = ExecutionMode.PRODUCTION,
        modify_pythoncommons_logger_names: bool = True,
        remove_existing_handlers: bool = True,
        sanity_check_number_of_handlers: bool = True,
        disable_propagation: bool = True,
        enable_logging_setup_debug_details: bool = False,
    ) -> SimpleLoggingSetupConfig:
        conf = SimpleLoggingSetupInputConfig(
            project_name,
            logger_name_prefix,
            debug,
            console_debug,
            format_str,
            file_postfix,
            execution_mode,
            modify_pythoncommons_logger_names,
            remove_existing_handlers,
            disable_propagation,
            enable_logging_setup_debug_details,
        )
        conf.specified_file_log_level = logging.DEBUG if debug else DEFAULT_LOG_LEVEL
        specified_file_log_level_name: str = logging.getLevelName(conf.specified_file_log_level)
        default_file_log_level_name: str = logging.getLevelName(DEFAULT_LOG_LEVEL)
        console_log_level: int = logging.DEBUG if debug else DEFAULT_LOG_LEVEL

        if not console_debug:
            console_log_level = DEFAULT_LOG_LEVEL

        final_format_str, formatter = SimpleLoggingSetup._determine_formatter(format_str)

        # This will init the root logger to the specified level
        logging.basicConfig(format=final_format_str, level=conf.specified_file_log_level)

        log_file_path_for_default_level = SimpleLoggingSetup._determine_log_file_path(
            default_file_log_level_name, file_postfix, execution_mode, project_name
        )
        log_file_path_for_specified_level = SimpleLoggingSetup._determine_log_file_path(
            specified_file_log_level_name, file_postfix, execution_mode, project_name
        )
        console_handler, conf.handlers, log_file_paths = SimpleLoggingSetup._determine_handlers(
            console_log_level,
            log_file_path_for_default_level,
            log_file_path_for_specified_level,
            conf.specified_file_log_level,
            formatter,
        )
        # Add fields to input config
        project_main_logger = SimpleLoggingSetup._setup_project_main_logger(conf)
        loggers: List[logging.Logger] = SimpleLoggingSetup.setup_existing_loggers(conf)

        logger_to_handler_count_dict = {logger.name: len(logger.handlers) for logger in loggers}
        project_main_logger.debug("Number of handlers on existing loggers: %s", logger_to_handler_count_dict)
        if sanity_check_number_of_handlers:
            SimpleLoggingSetup._sanity_check_number_of_handlers(conf, loggers)

        if enable_logging_setup_debug_details:
            logger = project_main_logger
            project_main_logger.info(
                "Resetting log level on logger: %s, as initial logging setup has been completed.", logger
            )
            SimpleLoggingSetup._set_level_on_logger(conf.specified_file_log_level, logger, logger)

        config = SimpleLoggingSetupConfig(
            project_name=project_name,
            execution_mode=execution_mode,
            file_log_level_name=specified_file_log_level_name,
            console_log_level_name=logging.getLevelName(console_log_level),
            formatter=final_format_str,
            console_handler=console_handler,
            file_handlers=list(filter(lambda h: isinstance(h, logging.FileHandler), conf.handlers)),
            main_project_logger=project_main_logger,
            log_file_paths=log_file_paths,
        )
        return config

    @staticmethod
    def _determine_handlers(
        console_log_level,
        log_file_path_for_default_level,
        log_file_path_for_specified_level,
        specified_file_log_level,
        formatter,
    ):
        log_file_paths: Dict[int, str] = {DEFAULT_LOG_LEVEL: log_file_path_for_default_level}
        file_handler = SimpleLoggingSetup._create_file_handler(log_file_path_for_default_level, DEFAULT_LOG_LEVEL)
        console_handler = None
        if PyTestUtils.is_pytest_execution():
            # IMPORTANT!
            # PyTest has its own handlers so a console handler wouldn't log anything.
            # Keep all handlers of PyTest and don't use the simple StreamHandler for the stdout stream.
            # An example list of these handlers:
            # 0 = {_LiveLoggingStreamHandler} <_LiveLoggingStreamHandler (DEBUG)>
            # 1 = {_FileHandler} <_FileHandler /<path>/pytest-logs.txt (DEBUG)>
            # 2 = {LogCaptureHandler} <LogCaptureHandler (DEBUG)>
            # 3 = {LogCaptureHandler} <LogCaptureHandler (DEBUG)>
            # More info here:
            # 1. https://stackoverflow.com/a/51633600/1106893
            # 2. https://docs.pytest.org/en/latest/how-to/logging.html#live-logs
            root_logger_handlers = logging.getLogger().handlers
            for rh in root_logger_handlers:
                if isinstance(rh, _LiveLoggingStreamHandler) and isinstance(rh.stream, TerminalReporter):
                    console_handler = rh
            if not console_handler:
                raise ValueError("Console handler not found among PyTest's handlers: {}".format(root_logger_handlers))
            handlers = [*root_logger_handlers, file_handler]
        else:
            console_handler = SimpleLoggingSetup._create_console_handler(console_log_level)
            handlers = [
                console_handler,
                file_handler,
            ]

        # Only add a second file handler if default logging level is different from specified.
        # Example: Default is logging.INFO, specified is logging.DEBUG
        if specified_file_log_level != DEFAULT_LOG_LEVEL:
            handler = SimpleLoggingSetup._create_file_handler(
                log_file_path_for_specified_level, specified_file_log_level
            )
            handlers.append(handler)
            log_file_paths[specified_file_log_level] = log_file_path_for_specified_level
        SimpleLoggingSetup._ALL_LOG_FILES.extend(log_file_paths.values())

        for h in handlers:
            h.setFormatter(formatter)
        return console_handler, handlers, log_file_paths

    @staticmethod
    def _determine_formatter(format_str):
        final_format_str = DEFAULT_FORMAT
        if format_str:
            final_format_str = format_str
        formatter = logging.Formatter(final_format_str)
        return final_format_str, formatter

    @staticmethod
    def _sanity_check_number_of_handlers(conf, loggers):
        wrong_number_of_handlers: Dict[str, List[logging.Handler]] = {}
        expected_no_of_handlers = len(conf.handlers)
        for logger in loggers:
            logger_name = logger.name
            if logger_name.startswith(PYTHONCOMMONS_PROJECT_NAME) or logger_name.startswith(conf.logger_name_prefix):
                if len(logger.handlers) != expected_no_of_handlers:
                    wrong_number_of_handlers[logger_name] = logger.handlers
        if len(wrong_number_of_handlers) > 0:
            raise ValueError(
                "Unexpected number of handlers on loggers. "
                f"Expected: {expected_no_of_handlers}, "
                f"Expected handlers: {conf.handlers}, "
                f"These loggers are having wrong number of handlers: {wrong_number_of_handlers}"
            )

    @staticmethod
    def _determine_log_file_path(file_log_level_name, file_postfix, exec_mode, project_name):
        if exec_mode == ExecutionMode.PRODUCTION:
            log_file_path = ProjectUtils.get_default_log_file(
                project_name, postfix=file_postfix, level_name=file_log_level_name
            )
        elif exec_mode == ExecutionMode.TEST:
            log_file_path = ProjectUtils.get_default_test_log_file(
                project_name, postfix=file_postfix, level_name=file_log_level_name
            )
        else:
            raise ValueError(f"Unknown execution mode: {exec_mode}")
        # else:
        #     final_log_dir = os.path.join(os.path.curdir, 'logs')
        #     final_log_file_path = SimpleLoggingSetup.get_default_log_file_path(file_log_level_name, final_log_dir,
        #                                                                        project_name)
        return log_file_path

    @staticmethod
    def get_default_log_file_path(level_str, log_dir, project_name):
        FileUtils.ensure_dir_created(log_dir)
        # Example file name: <project>-info-2020_03_12_01_19_48
        timestamp = _get_timestamp_as_str()
        logfilename = project_name
        if level_str:
            logfilename += "-" + level_str
        logfilename += "-" + timestamp + ".log"
        log_file = os.path.join(log_dir, logfilename)
        return log_file

    @staticmethod
    def _create_console_handler(level):
        ch = logging.StreamHandler(stream=DEFAULT_CONSOLE_STREAM)
        ch.setLevel(level)
        return ch

    @staticmethod
    def _create_file_handler(log_file_path, level: int):
        fh = TimedRotatingFileHandler(log_file_path, when="midnight")
        fh.suffix = "%Y_%m_%d.log"
        fh.setLevel(level)
        return fh

    @staticmethod
    def _setup_project_main_logger(conf: SimpleLoggingSetupInputConfig):
        logger = logging.getLogger(conf.logger_name_prefix)
        logger.propagate = False
        SimpleLoggingSetup._set_level_and_add_handlers(conf, logger)
        return logger

    @staticmethod
    def setup_existing_loggers(conf: SimpleLoggingSetupInputConfig):
        level = conf.specified_file_log_level
        level_name: str = logging.getLevelName(level)
        logger_names, loggers = SimpleLoggingSetup.get_all_loggers_from_loggerdict(conf.project_main_logger)
        project_specific_loggers = SimpleLoggingSetup.get_project_specific_loggers(
            loggers, conf.logger_name_prefix, append_dot=True
        )

        if not project_specific_loggers:
            print(
                "Cannot find any project specific loggers with project name '%s', found loggers: %s",
                conf.logger_name_prefix,
                logger_names,
            )
        else:
            conf.project_main_logger.debug(
                "Setting logging level to '%s' on the following project-specific loggers: %s.", level_name, logger_names
            )
        SimpleLoggingSetup._set_level_and_add_handlers_on_loggers(conf, project_specific_loggers, logger_names)

        # Add handlers for non-project specific loggers as well (Git, GoogleApiWrapper, etc.)
        except_prefixes = [conf.logger_name_prefix]
        if not conf.modify_pythoncommons_logger_names:
            except_prefixes.append(PYTHONCOMMONS_PROJECT_NAME)
        non_project_specific_loggers = SimpleLoggingSetup.get_loggers(loggers, except_prefixes)
        SimpleLoggingSetup._set_level_and_add_handlers_on_loggers(conf, non_project_specific_loggers, logger_names)
        return loggers

    @staticmethod
    def get_all_loggers_from_loggerdict(logger=None):
        loggers: List[logging.Logger] = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
        logger_names: List[str] = list(map(lambda x: x.name, loggers))
        if logger:
            logger.debug("Discovered loggers: %s", logger_names)
        return logger_names, loggers

    @staticmethod
    def get_project_specific_loggers(loggers, logger_name_prefix, append_dot=False):
        if append_dot:
            logger_name_prefix = logger_name_prefix + "."
        loggers: List[logging.Logger] = list(filter(lambda x: x.name.startswith(logger_name_prefix), loggers))
        return loggers

    @staticmethod
    def get_loggers(loggers, except_prefixes: List[str], append_dots=False):
        if not except_prefixes:
            return loggers
        if append_dots:
            except_prefixes = list(map(lambda e: e + "." if e[-1] != "." else e, except_prefixes))

        filtered_loggers = []
        for logger in loggers:
            for prefix in except_prefixes:
                if not logger.name.startswith(prefix):
                    filtered_loggers.append(logger)
        return filtered_loggers

    @staticmethod
    def get_pythoncommons_loggers(loggers):
        pythoncommons_loggers = list(filter(lambda x: x.name.startswith(PYTHONCOMMONS_PROJECT_NAME + "."), loggers))
        return pythoncommons_loggers

    @staticmethod
    def _set_level_and_add_handlers_on_loggers(
        conf: SimpleLoggingSetupInputConfig, loggers: List[logging.Logger], logger_names: List[str], recursive=True
    ):
        level = conf.specified_file_log_level
        level_name = logging.getLevelName(level)
        for logger in loggers:
            SimpleLoggingSetup._set_level_and_add_handlers(conf, logger, recursive=recursive)
            if conf.disable_propagation and logger.propagate:
                conf.project_main_logger.debug("Disabling propagate on logger: %s", logger.name)
                logger.propagate = False
        conf.project_main_logger.info("Set level to '%s' on these discovered loggers: %s", level_name, logger_names)

    @staticmethod
    def _set_level_and_add_handlers(conf: SimpleLoggingSetupInputConfig, logger: logging.Logger, recursive=True):
        level = conf.specified_file_log_level
        existing_handlers = logger.handlers
        # has_console_handler = len(list(filter(lambda h: SimpleLoggingSetup._is_console_handler(h), existing_handlers)))
        # has_file_handler = len(list(filter(lambda h: SimpleLoggingSetup._is_file_handler(h), existing_handlers)))
        is_project_main_logger = SimpleLoggingSetup._is_project_main_logger(conf, logger)
        main_logger = conf.project_main_logger
        if is_project_main_logger:
            conf.project_main_logger = logger
            main_logger = conf.project_main_logger
            if not conf.enable_logging_setup_debug_details:
                main_logger.info("Disabling debug logs of initial LoggingSetup. Logger: %s", main_logger)
                main_logger.setLevel(DEFAULT_LOG_LEVEL)

        # If we are in TEST mode and the handler is a FileHandler, don't replace it
        callback = (
            lambda handler: conf.execution_mode == ExecutionMode.TEST
            and not SimpleLoggingSetup._is_file_handler(handler)
        )

        if conf.remove_existing_handlers:
            main_logger.debug("Removing existing handlers from logger: %s, handlers: %s", logger, existing_handlers)

            # Handle project main logger specially
            if is_project_main_logger:
                if PyTestUtils.is_pytest_execution():
                    # Simply remove & add all handlers
                    SimpleLoggingSetup._remove_handlers_from_logger(main_logger, logger, type=None)
                    SimpleLoggingSetup._add_handlers_to_logger(main_logger, logger, conf.handlers)
                else:
                    SimpleLoggingSetup._remove_handlers_from_logger(main_logger, logger, type=HandlerType.CONSOLE)
                    SimpleLoggingSetup._add_handlers_to_logger(
                        main_logger, logger, conf.handlers, type=HandlerType.CONSOLE
                    )
                    SimpleLoggingSetup._remove_handlers_from_logger(main_logger, logger, type=HandlerType.FILE)
                    SimpleLoggingSetup._add_handlers_to_logger(
                        main_logger, logger, conf.handlers, type=HandlerType.FILE
                    )
                return
            else:
                SimpleLoggingSetup._remove_handlers_from_logger(main_logger, logger, callback=callback)

        if not is_project_main_logger or conf.enable_logging_setup_debug_details:
            SimpleLoggingSetup._set_level_on_logger(level, logger, main_logger)
        kwargs = {}
        if conf.remove_existing_handlers and len(logger.handlers) != 0:
            # If we selectively removed handlers with the callback,
            # we also want to selectively add handlers with the same callback
            kwargs["callback"] = callback

        SimpleLoggingSetup._add_handlers_to_logger(main_logger, logger, conf.handlers, **kwargs)
        if logger.parent:
            SimpleLoggingSetup._set_level_and_add_handlers(conf, logger.parent, recursive=recursive)

    @staticmethod
    def _set_level_on_logger(level, logger, project_main_logger):
        project_main_logger.debug("Setting log level to '%s' on logger '%s'", logging.getLevelName(level), logger)
        logger.setLevel(level)

    @staticmethod
    def _set_level_on_handler(level, handler, project_main_logger=None):
        project_main_logger.debug("Setting log level to '%s' on handler '%s'", logging.getLevelName(level), handler)
        handler.setLevel(level)

    @staticmethod
    def _add_handlers_to_logger(
        project_main_logger, logger, handlers, type: HandlerType = None, callback: Callable = None
    ):
        filtered_handlers = SimpleLoggingSetup._filter_handlers_by_type(handlers, type)
        for handler in filtered_handlers:
            if (callback and callback(handler)) or not callback:
                project_main_logger.debug("Adding handler '%s' to logger '%s'", handler, logger)
                logger.addHandler(handler)

    @staticmethod
    def _remove_handlers_from_logger(project_main_logger, logger, type: HandlerType = None, callback: Callable = None):
        existing_handlers = logger.handlers
        filtered_handlers = SimpleLoggingSetup._filter_handlers_by_type(existing_handlers, type)
        for handler in filtered_handlers:
            if (callback and callback(handler)) or not callback:
                project_main_logger.debug("Removing handler '%s' from logger '%s'", handler, logger)
                logger.removeHandler(handler)

    @staticmethod
    def _same_loggers(l1, l2):
        return l1.name == l2.name

    @staticmethod
    def _is_project_main_logger(conf, logger):
        return logger.name == conf.logger_name_prefix

    @staticmethod
    def _is_console_handler(handler):
        return isinstance(handler, logging.StreamHandler)

    @staticmethod
    def _is_file_handler(handler):
        return isinstance(handler, logging.FileHandler)

    @staticmethod
    def _filter_handlers_by_type(handlers, type: HandlerType = None):
        if not type:
            return handlers
        if type == HandlerType.FILE:
            return list(filter(lambda h: isinstance(h, logging.FileHandler), handlers))
        elif type == HandlerType.CONSOLE:
            return list(
                filter(lambda h: isinstance(h, logging.StreamHandler) and h.stream == DEFAULT_CONSOLE_STREAM, handlers)
            )

    @staticmethod
    def create_command_logger(name: str):
        cmd_logger = logging.getLogger(name)
        cmd_log_handler = logging.StreamHandler(stream=sys.stdout)
        cmd_logger.propagate = False
        cmd_logger.addHandler(cmd_log_handler)
        cmd_log_handler.setFormatter(logging.Formatter("%(message)s"))
        return cmd_logger

    @classmethod
    def get_all_log_files(cls):
        return cls._ALL_LOG_FILES
